"""Storing the fixture to be used for the tests. Note that this file will only contain
fixture that are made directly from parsing the files. Any additional operations like
concat should be done at local level."""

import pytest
from pytest import LogCaptureFixture
from alchemtest.amber import load_bace_example, load_simplesolvated, load_tyk2_example
from alchemtest.gmx import (
    load_benzene,
    load_ethanol,
    load_expanded_ensemble_case_1,
    load_expanded_ensemble_case_2,
    load_expanded_ensemble_case_3,
    load_water_particle_with_total_energy,
    load_water_particle_with_potential_energy,
    load_water_particle_without_energy,
    load_ABFE,
)
import pandas as pd
from alchemtest.gomc import load_benzene as gomc_load_benzene
from alchemtest.namd import (
    load_tyr2ala,
    load_idws,
    load_restarted,
    load_restarted_reversed,
)
from loguru import logger

from alchemlyb.parsing import gmx, amber, gomc, namd


@pytest.fixture
def gmx_benzene() -> dict[str, list[str]]:
    dataset = load_benzene()
    return dataset["data"]  # type: ignore[no-any-return]


@pytest.fixture
def gmx_benzene_Coulomb_dHdl(gmx_benzene: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_dHdl(file, T=300) for file in gmx_benzene["Coulomb"]]


@pytest.fixture
def gmx_benzene_VDW_dHdl(gmx_benzene: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_dHdl(file, T=300) for file in gmx_benzene["VDW"]]


@pytest.fixture
def gmx_ethanol() -> dict[str, list[str]]:
    dataset = load_ethanol()
    return dataset["data"]  # type: ignore[no-any-return]


@pytest.fixture
def gmx_ethanol_Coulomb_dHdl(gmx_ethanol: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_dHdl(file, T=300) for file in gmx_ethanol["Coulomb"]]


@pytest.fixture
def gmx_ethanol_VDW_dHdl(gmx_ethanol: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_dHdl(file, T=300) for file in gmx_ethanol["VDW"]]


@pytest.fixture
def gmx_benzene_Coulomb_u_nk(gmx_benzene: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_u_nk(file, T=300) for file in gmx_benzene["Coulomb"]]


@pytest.fixture
def gmx_benzene_VDW_u_nk(gmx_benzene: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_u_nk(file, T=300) for file in gmx_benzene["VDW"]]


@pytest.fixture
def gmx_ABFE() -> dict[str, list[str]]:
    dataset = load_ABFE()
    return dataset["data"]  # type: ignore[no-any-return]


@pytest.fixture
def gmx_ABFE_complex_u_nk(gmx_ABFE: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_u_nk(file, T=300) for file in gmx_ABFE["complex"]]


@pytest.fixture
def gmx_ABFE_complex_dHdl(gmx_ABFE: dict[str, list[str]]) -> list[pd.DataFrame]:
    return [gmx.extract_dHdl(file, T=300) for file in gmx_ABFE["complex"]]


@pytest.fixture
def gmx_expanded_ensemble_case_1() -> list[pd.DataFrame]:
    dataset = load_expanded_ensemble_case_1()

    return [
        gmx.extract_u_nk(filename, T=300, filter=False)
        for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_expanded_ensemble_case_1_dHdl() -> list[pd.DataFrame]:
    dataset = load_expanded_ensemble_case_1()

    return [
        gmx.extract_dHdl(filename, T=300, filter=False)
        for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_expanded_ensemble_case_2() -> list[pd.DataFrame]:
    dataset = load_expanded_ensemble_case_2()

    return [
        gmx.extract_u_nk(filename, T=300, filter=False)
        for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_expanded_ensemble_case_2_dHdl() -> list[pd.DataFrame]:
    dataset = load_expanded_ensemble_case_2()

    return [
        gmx.extract_dHdl(filename, T=300, filter=False)
        for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_expanded_ensemble_case_3() -> list[pd.DataFrame]:
    dataset = load_expanded_ensemble_case_3()

    return [
        gmx.extract_u_nk(filename, T=300, filter=False)
        for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_expanded_ensemble_case_3_dHdl() -> list[pd.DataFrame]:
    dataset = load_expanded_ensemble_case_3()

    return [
        gmx.extract_dHdl(filename, T=300, filter=False)
        for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_water_particle_with_total_energy() -> list[pd.DataFrame]:
    dataset = load_water_particle_with_total_energy()

    return [
        gmx.extract_u_nk(filename, T=300) for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_water_particle_with_total_energy_dHdl() -> list[pd.DataFrame]:
    dataset = load_water_particle_with_total_energy()

    return [
        gmx.extract_dHdl(filename, T=300) for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_water_particle_with_potential_energy() -> list[pd.DataFrame]:
    dataset = load_water_particle_with_potential_energy()

    return [
        gmx.extract_u_nk(filename, T=300) for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_water_particle_with_potential_energy_dHdl() -> list[pd.DataFrame]:
    dataset = load_water_particle_with_potential_energy()

    return [
        gmx.extract_dHdl(filename, T=300) for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_water_particle_without_energy() -> list[pd.DataFrame]:
    dataset = load_water_particle_without_energy()

    return [
        gmx.extract_u_nk(filename, T=300) for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def gmx_water_particle_without_energy_dHdl() -> list[pd.DataFrame]:
    dataset = load_water_particle_without_energy()

    return [
        gmx.extract_dHdl(filename, T=300) for filename in dataset["data"]["AllStates"]
    ]


@pytest.fixture
def amber_simplesolvated() -> dict[str, list[str]]:
    dataset = load_simplesolvated()
    return dataset["data"]  # type: ignore[no-any-return]


@pytest.fixture
def amber_simplesolvated_charge_dHdl(
    amber_simplesolvated: dict[str, list[str]],
) -> list[pd.DataFrame]:
    return [
        amber.extract_dHdl(filename, T=298.0)  # type: ignore[misc]
        for filename in amber_simplesolvated["charge"]
    ]


@pytest.fixture
def amber_simplesolvated_vdw_dHdl(
    amber_simplesolvated: dict[str, list[str]],
) -> list[pd.DataFrame]:
    return [
        amber.extract_dHdl(filename, T=298.0)  # type: ignore[misc]
        for filename in amber_simplesolvated["vdw"]
    ]


@pytest.fixture
def amber_bace_example_complex_vdw() -> list[pd.DataFrame]:
    dataset = load_bace_example()

    return [
        amber.extract_u_nk(filename, T=298.0)  # type: ignore[misc]
        for filename in dataset["data"]["complex"]["vdw"]
    ]


@pytest.fixture
def amber_tyk2_example_complex() -> list[pd.DataFrame]:
    dataset = load_tyk2_example()

    return [
        amber.extract_dHdl(filename, T=300.0)
        for filename in dataset["data"]["complex"]  # type: ignore[misc]
    ]


@pytest.fixture
def gomc_benzene() -> list[str]:
    dataset = gomc_load_benzene()
    return dataset["data"]  # type: ignore[no-any-return]


@pytest.fixture
def gomc_benzene_u_nk(gomc_benzene: list[str]) -> list[pd.DataFrame]:
    return [gomc.extract_u_nk(filename, T=298) for filename in gomc_benzene]


@pytest.fixture
def gomc_benzene_dHdl(gomc_benzene: list[str]) -> list[pd.DataFrame]:
    return [gomc.extract_dHdl(filename, T=298) for filename in gomc_benzene]


@pytest.fixture
def namd_tyr2ala() -> pd.DataFrame:
    dataset = load_tyr2ala()
    u_nk1 = namd.extract_u_nk(dataset["data"]["forward"][0], T=300)
    u_nk2 = namd.extract_u_nk(dataset["data"]["backward"][0], T=300)

    # combine dataframes of fwd and rev directions
    u_nk1[u_nk1.isna()] = u_nk2
    u_nk = u_nk1.sort_index(level=u_nk1.index.names[1:])

    return u_nk  # type: ignore[no-any-return]


@pytest.fixture
def namd_idws() -> pd.DataFrame:
    dataset = load_idws()
    u_nk = namd.extract_u_nk(dataset["data"]["forward"], T=300)

    return u_nk  # type: ignore[no-any-return]


@pytest.fixture
def namd_idws_restarted() -> pd.DataFrame:
    dataset = load_restarted()
    u_nk = namd.extract_u_nk(dataset["data"]["both"], T=300)

    return u_nk  # type: ignore[no-any-return]


@pytest.fixture
def namd_idws_restarted_reversed() -> pd.DataFrame:
    dataset = load_restarted_reversed()
    u_nk = namd.extract_u_nk(dataset["data"]["both"], T=300)

    return u_nk  # type: ignore[no-any-return]


@pytest.fixture
def caplog(caplog: LogCaptureFixture) -> LogCaptureFixture:  # type: ignore[misc]
    handler_id = logger.add(caplog.handler, format="{message}")
    yield caplog
    logger.remove(handler_id)



import os

import numpy as np
import pandas as pd
import pytest
from alchemtest.amber import load_bace_example
from alchemtest.gmx import load_ABFE
import joblib

import alchemlyb.parsing.amber
from alchemlyb.workflows.abfe import ABFE


@pytest.fixture(scope="module")
def workflow(tmp_path_factory):
    outdir = tmp_path_factory.mktemp("out")
    dir = os.path.dirname(load_ABFE()["data"]["complex"][0])
    workflow = ABFE(
        units="kcal/mol",
        software="GROMACS",
        dir=dir,
        prefix="dhdl",
        suffix="xvg",
        T=310,
        outdirectory=str(outdir),
    )
    workflow.run(
        skiptime=10,
        uncorr="dE",
        threshold=50,
        estimators=("MBAR", "BAR", "TI"),
        overlap="O_MBAR.pdf",
        breakdown=True,
        forwrev=10,
    )
    return workflow


class TestInit:
    def test_nofilematch(self, tmp_path):
        with pytest.raises(ValueError, match="No file has been matched to"):
            ABFE(
                dir=str(tmp_path),
                prefix="dhdl",
                suffix="xvg",
                T=310,
            )

    def test_notdir(self):
        with pytest.raises(ValueError, match="The input directory `dir`="):
            ABFE(
                dir="abfasfsd",
                prefix="dhdl",
                suffix="xvg",
                T=310,
            )

    def test_wildcard_in_dir(self):
        with pytest.raises(ValueError):
            with pytest.warns(match="A real directory is expected in `dir`="):
                ABFE(
                    dir="/*/",
                    prefix="dhdl",
                    suffix="xvg",
                    T=310,
                )


class TestRun:
    def test_none(self, workflow):
        """Don't run anything"""
        workflow.run(
            uncorr=None, estimators=None, overlap=None, breakdown=None, forwrev=None
        )

    def test_invalid_estimator(self, workflow):
        with pytest.raises(ValueError, match=r"Estimator aaa is not supported."):
            workflow.run(
                uncorr=None,
                estimators="aaa",
                overlap=None,
                breakdown=None,
                forwrev=None,
            )

    def test_single_estimator(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "u_nk_list", [])
        monkeypatch.setattr(workflow, "dHdl_list", [])
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])
        monkeypatch.setattr(workflow, "estimator", dict())
        workflow.run(
            uncorr=None,
            estimators="MBAR",
            overlap=None,
            breakdown=True,
            forwrev=None,
        )
        assert "MBAR" in workflow.estimator

    @pytest.mark.parametrize("forwrev", [None, False, 0])
    def test_no_forwrev(self, workflow, monkeypatch, forwrev):
        monkeypatch.setattr(workflow, "convergence", None)
        workflow.run(
            uncorr=None,
            estimators=None,
            overlap=None,
            breakdown=None,
            forwrev=forwrev,
        )
        assert workflow.convergence is None


class TestRead:
    def test_default(self, workflow):
        """test if the files has been loaded correctly."""
        assert len(workflow.u_nk_list) == 30
        assert len(workflow.dHdl_list) == 30
        assert all([len(u_nk) == 1001 for u_nk in workflow.u_nk_list])
        assert all([len(dHdl) == 1001 for dHdl in workflow.dHdl_list])

    def test_no_parser(self):
        dir = os.path.dirname(load_ABFE()["data"]["complex"][0])
        with pytest.raises(NotImplementedError):
            ABFE(
                units="kcal/mol",
                software="aaa",
                dir=dir,
                prefix="dhdl",
                suffix="xvg",
                T=310,
            )

    @pytest.mark.parametrize("read_u_nk", [True, False])
    @pytest.mark.parametrize("read_dHdl", [True, False])
    def test_read_TI_FEP(self, workflow, monkeypatch, read_u_nk, read_dHdl):
        monkeypatch.setattr(workflow, "u_nk_list", [])
        monkeypatch.setattr(workflow, "dHdl_list", [])
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])
        workflow.read(read_u_nk, read_dHdl)
        if read_u_nk:
            assert len(workflow.u_nk_list) == 30
        else:
            assert len(workflow.u_nk_list) == 0

        if read_dHdl:
            assert len(workflow.dHdl_list) == 30
        else:
            assert len(workflow.dHdl_list) == 0

    def test_read_invalid_u_nk(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])

        def extract_u_nk(self, T):
            raise IOError("Error read u_nk.")

        monkeypatch.setattr(workflow, "_extract_u_nk", extract_u_nk)
        with pytest.raises(OSError, match=r"Error reading u_nk"):
            workflow.read()

    def test_read_invalid_dHdl(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])

        def extract_dHdl(self, T):
            raise IOError("Error read dHdl.")

        monkeypatch.setattr(workflow, "_extract_dHdl", extract_dHdl)
        with pytest.raises(OSError, match=r"Error reading dHdl"):
            workflow.read()


class TestSubsample:
    def test_default(self, workflow):
        """Test if the data has been shrinked by subsampling."""
        assert len(workflow.u_nk_sample_list) == 30
        assert len(workflow.dHdl_sample_list) == 30
        assert all([len(u_nk) < 1001 for u_nk in workflow.u_nk_sample_list])
        assert all([len(dHdl) < 1001 for dHdl in workflow.dHdl_sample_list])

    def test_uncorr_threshold(self, workflow, monkeypatch):
        """Test if the full data will be used when the number of data points
        are less than the threshold."""
        monkeypatch.setattr(
            workflow, "u_nk_list", [u_nk[:40] for u_nk in workflow.u_nk_list]
        )
        monkeypatch.setattr(
            workflow, "dHdl_list", [dHdl[:40] for dHdl in workflow.dHdl_list]
        )
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])
        workflow.preprocess(threshold=50)
        assert all([len(u_nk) == 40 for u_nk in workflow.u_nk_sample_list])
        assert all([len(dHdl) == 40 for dHdl in workflow.dHdl_sample_list])

    def test_no_u_nk_preprocess(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "u_nk_list", [])
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])
        workflow.preprocess(threshold=50)
        assert len(workflow.u_nk_list) == 0

    def test_no_dHdl_preprocess(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "dHdl_list", [])
        monkeypatch.setattr(workflow, "u_nk_sample_list", [])
        monkeypatch.setattr(workflow, "dHdl_sample_list", [])
        workflow.preprocess(threshold=50)
        assert len(workflow.dHdl_list) == 0


class TestEstimator:
    def test_default(self, workflow):
        """Test if all three estimators have been used."""
        assert len(workflow.estimator) == 3
        assert "MBAR" in workflow.estimator
        assert "TI" in workflow.estimator
        assert "BAR" in workflow.estimator

    def test_method(self, workflow, monkeypatch):
        """Test if the method keyword could be passed to the AutoMBAR estimator."""
        monkeypatch.setattr(workflow, "estimator", dict())
        workflow.estimate(estimators="MBAR", method="adaptive")
        assert "MBAR" in workflow.estimator

    def test_summary(self, workflow):
        """Test if if the summary is right."""
        summary = workflow.generate_result()
        assert np.isclose(summary["MBAR"]["Stages"]["TOTAL"], 21.8, 0.1)
        assert np.isclose(summary["TI"]["Stages"]["TOTAL"], 21.8, 0.1)
        assert np.isclose(summary["BAR"]["Stages"]["TOTAL"], 21.8, 0.1)

    def test_no_name_estimate(self, workflow):
        with pytest.raises(ValueError):
            workflow.estimate("aaa")

    def test_single_estimator_mbar(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "estimator", dict())
        workflow.estimate(estimators="MBAR")
        assert len(workflow.estimator) == 1
        assert "MBAR" in workflow.estimator
        summary = workflow.generate_result()
        assert np.isclose(summary["MBAR"]["Stages"]["TOTAL"], 21.645742066696315, 0.1)

    def test_mbar_n_bootstraps(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "estimator", dict())
        workflow.estimate(estimators="MBAR", n_bootstraps=2)
        summary = workflow.generate_result()
        bootstrap_error = summary["MBAR_Error"]["Stages"]["TOTAL"]
        monkeypatch.setattr(workflow, "estimator", dict())
        workflow.estimate(estimators="MBAR", n_bootstraps=0)
        summary = workflow.generate_result()
        non_bootstrap_error = summary["MBAR_Error"]["Stages"]["TOTAL"]
        assert bootstrap_error != non_bootstrap_error

    def test_single_estimator_ti(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "estimator", dict())
        monkeypatch.setattr(workflow, "summary", None)
        workflow.estimate(estimators="TI")
        summary = workflow.generate_result()
        assert np.isclose(summary["TI"]["Stages"]["TOTAL"], 21.65, 0.1)

    def test_unprocessed_u_nk(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "u_nk_sample_list", None)
        monkeypatch.setattr(workflow, "estimator", dict())
        workflow.estimate()
        assert len(workflow.estimator) == 3
        assert "MBAR" in workflow.estimator

    def test_unpertubed_lambda(self, workflow, monkeypatch, gmx_benzene_Coulomb_dHdl):
        """Test the if two lamdas present and one of them is not pertubed.

                                               fep  bound
        time    fep-lambda bound-lambda
        0.0     0.5        0             12.958159      0
        10.0    0.5        0             -1.062968      0
        20.0    0.5        0              1.019020      0
        30.0    0.5        0              5.029051      0
        40.0    0.5        0              7.768072      0

        Where only fep-lambda changes but the bonded-lambda is always 0.
        """
        monkeypatch.setattr(workflow, "u_nk_list", [])
        monkeypatch.setattr(workflow, "u_nk_sample_list", None)
        monkeypatch.setattr(workflow, "dHdl_list", gmx_benzene_Coulomb_dHdl)
        monkeypatch.setattr(workflow, "dHdl_sample_list", gmx_benzene_Coulomb_dHdl)
        monkeypatch.setattr(workflow, "estimator", dict())
        monkeypatch.setattr(workflow, "summary", None)
        # Add another lambda column
        for dHdl in workflow.dHdl_sample_list:
            dHdl.insert(1, "bound-lambda", [1.0] * len(dHdl))
            dHdl.insert(1, "bound", [1.0] * len(dHdl))
            dHdl.set_index("bound-lambda", append=True, inplace=True)

        workflow.estimate(estimators="TI")
        summary = workflow.generate_result()
        assert np.isclose(summary["TI"]["Stages"]["bound"], 0)


class TestVisualisation:
    def test_plot_O_MBAR(self, workflow):
        """test if the O_MBAR.pdf has been plotted."""
        assert os.path.isfile(os.path.join(workflow.out, "O_MBAR.pdf"))

    def test_plot_dhdl_TI(self, workflow):
        """test if the dhdl_TI.pdf has been plotted."""
        assert os.path.isfile(os.path.join(workflow.out, "dhdl_TI.pdf"))

    def test_plot_dF_state(self, workflow):
        """test if the dF_state.pdf has been plotted."""
        assert os.path.isfile(os.path.join(workflow.out, "dF_state.pdf"))
        assert os.path.isfile(os.path.join(workflow.out, "dF_state_long.pdf"))

    def test_dhdl_TI_noTI(self, workflow, monkeypatch):
        """Test to plot the dhdl_TI when ti estimator is not there"""
        no_TI = workflow.estimator.copy()
        no_TI.pop("TI")
        monkeypatch.setattr(workflow, "estimator", no_TI)
        with pytest.raises(ValueError):
            workflow.plot_ti_dhdl(dhdl_TI="dhdl_TI.pdf")

    def test_noMBAR_for_plot_overlap_matrix(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "estimator", {})
        assert workflow.plot_overlap_matrix() is None


class TestConvergence:
    def test_default(self, workflow):
        """test if the dF_state.pdf has been plotted."""
        assert os.path.isfile(os.path.join(workflow.out, "dF_t.pdf"))
        assert len(workflow.convergence) == 10

    def test_method(self, workflow, monkeypatch):
        """Test if the method keyword could be passed to the AutoMBAR estimator from convergence."""
        monkeypatch.setattr(workflow, "convergence", None)
        workflow.check_convergence(2, estimator="MBAR", method="adaptive")
        assert len(workflow.convergence) == 2

    def test_nosample_u_nk(self, workflow, monkeypatch):
        """test if the convergence routine would use the unsampled data
        when the data has not been subsampled."""
        monkeypatch.setattr(workflow, "u_nk_sample_list", None)
        monkeypatch.setattr(workflow, "convergence", None)
        workflow.check_convergence(10)
        assert len(workflow.convergence) == 10

    def test_no_u_nk_for_check_convergence(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "u_nk_list", None)
        monkeypatch.setattr(workflow, "u_nk_sample_list", None)
        with pytest.raises(ValueError):
            workflow.check_convergence(10, estimator="MBAR")

    def test_no_dHdl_for_check_convergence(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "dHdl_list", None)
        monkeypatch.setattr(workflow, "dHdl_sample_list", None)
        with pytest.raises(ValueError):
            workflow.check_convergence(10, estimator="TI")

    def test_bar_convergence(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "convergence", None)
        workflow.check_convergence(10, estimator="BAR")
        assert len(workflow.convergence) == 10

    def test_convergence_invalid_estimator(self, workflow):
        with pytest.raises(ValueError):
            workflow.check_convergence(10, estimator="aaa")

    def test_ti_convergence(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "convergence", None)
        workflow.check_convergence(10, estimator="TI")
        assert len(workflow.convergence) == 10

    def test_preserve_unit(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "convergence", None)
        monkeypatch.setattr(workflow, "units", "kcal/mol")
        workflow.check_convergence(2, estimator="TI")
        assert np.allclose(workflow.convergence["data_fraction"], [0.5, 1.0])

    def test_unprocessed_dhdl(self, workflow, monkeypatch):
        monkeypatch.setattr(workflow, "dHdl_sample_list", None)
        monkeypatch.setattr(workflow, "convergence", None)
        workflow.check_convergence(10, estimator="TI")
        assert len(workflow.convergence) == 10


class TestUnit:
    def test_no_update_units(self, workflow):
        assert workflow.update_units() is None


class Test_automatic_amber:
    """Test the full automatic workflow for load_ABFE from alchemtest.amber for
    three stage transformation."""

    @staticmethod
    @pytest.fixture(scope="class")
    def workflow(tmp_path_factory):
        outdir = tmp_path_factory.mktemp("out")
        dir, _ = os.path.split(
            os.path.dirname(load_bace_example()["data"]["complex"]["vdw"][0])
        )

        workflow = ABFE(
            units="kcal/mol",
            software="AMBER",
            dir=dir,
            prefix="ti",
            suffix="bz2",
            T=298.0,
            outdirectory=str(outdir),
        )
        workflow.read()
        workflow.estimate(estimators="TI")
        return workflow

    def test_summary(self, workflow):
        """Test if if the summary is right."""
        summary = workflow.generate_result()
        assert np.isclose(summary["TI"]["Stages"]["TOTAL"], 1.40405980473, 0.1)


class Test_automatic_parquet:
    """Test the full automatic workflow for load_ABFE from parquet data."""

    @staticmethod
    @pytest.fixture(scope="class")
    def workflow(tmp_path_factory):
        outdir = tmp_path_factory.mktemp("out")
        for i, u_nk in enumerate(load_bace_example()["data"]["complex"]["vdw"]):
            df = alchemlyb.parsing.amber.extract_u_nk(u_nk, T=298)
            df.to_parquet(path=f"{outdir}/u_nk_{i}.parquet", index=True)

        workflow = ABFE(
            units="kcal/mol",
            software="PARQUET",
            dir=str(outdir),
            prefix="u_nk_",
            suffix="parquet",
            T=298.0,
            outdirectory=str(outdir),
        )
        workflow.read()
        workflow.estimate(estimators="BAR")
        return workflow

    def test_summary(self, workflow):
        """Test if if the summary is right."""
        summary = workflow.generate_result()
        assert np.isclose(summary["BAR"]["Stages"]["TOTAL"], 1.40405980473, 0.1)


class TestParallel:
    @pytest.fixture(scope="class")
    def workflow(self, tmp_path_factory):
        outdir = tmp_path_factory.mktemp("out")
        (outdir / "dhdl_00.xvg").symlink_to(load_ABFE()["data"]["complex"][0])
        (outdir / "dhdl_01.xvg").symlink_to(load_ABFE()["data"]["complex"][1])
        workflow = ABFE(
            units="kcal/mol",
            software="GROMACS",
            dir=str(outdir),
            prefix="dhdl",
            suffix="xvg",
            T=310,
        )
        workflow.read()
        workflow.preprocess()
        return workflow

    @pytest.fixture(scope="class")
    def parallel_workflow(self, tmp_path_factory):
        outdir = tmp_path_factory.mktemp("out")
        (outdir / "dhdl_00.xvg").symlink_to(load_ABFE()["data"]["complex"][0])
        (outdir / "dhdl_01.xvg").symlink_to(load_ABFE()["data"]["complex"][1])
        workflow = ABFE(
            units="kcal/mol",
            software="GROMACS",
            dir=str(outdir),
            prefix="dhdl",
            suffix="xvg",
            T=310,
        )
        with joblib.parallel_config(backend="threading"):
            # The default backend is "loky", which is more robust but somehow didn't
            # play well with pytest, but "loky" is perfectly fine outside pytest.
            workflow.read(n_jobs=2)
            workflow.preprocess(n_jobs=2)
        return workflow

    def test_read(self, workflow, parallel_workflow):
        pd.testing.assert_frame_equal(
            workflow.u_nk_list[0], parallel_workflow.u_nk_list[0]
        )
        pd.testing.assert_frame_equal(
            workflow.u_nk_list[1], parallel_workflow.u_nk_list[1]
        )

    def test_preprocess(self, workflow, parallel_workflow):
        pd.testing.assert_frame_equal(
            workflow.u_nk_sample_list[0], parallel_workflow.u_nk_sample_list[0]
        )
        pd.testing.assert_frame_equal(
            workflow.u_nk_sample_list[1], parallel_workflow.u_nk_sample_list[1]
        )



"""GOMC parser tests."""

from alchemtest.gomc import load_benzene

from alchemlyb.parsing.gomc import extract_dHdl, extract_u_nk, extract


def test_dHdl():
    """Test that dHdl has the correct form when extracted from files."""
    dataset = load_benzene()

    for filename in dataset["data"]:
        dHdl = extract_dHdl(filename, T=298)

        assert dHdl.index.names == ["time", "Coulomb-lambda", "VDW-lambda"]
        assert dHdl.shape == (1000, 2)


def test_u_nk():
    """Test that u_nk has the correct form when extracted from files."""
    dataset = load_benzene()

    for filename in dataset["data"]:
        u_nk = extract_u_nk(filename, T=298)

        assert u_nk.index.names == ["time", "Coulomb-lambda", "VDW-lambda"]
        assert u_nk.shape == (1000, 23)


def test_extract():
    """Test that u_nk and dHdl have the correct form when extracted from files."""
    dataset = load_benzene()

    df_dict = extract(dataset["data"][0], T=298)

    assert df_dict["u_nk"].index.names == ["time", "Coulomb-lambda", "VDW-lambda"]
    assert df_dict["u_nk"].shape == (1000, 23)
    assert df_dict["dHdl"].index.names == ["time", "Coulomb-lambda", "VDW-lambda"]
    assert df_dict["dHdl"].shape == (1000, 2)



"""Tests for all TI-based estimators in ``alchemlyb``."""

import pandas as pd
import numpy as np
import copy
import pytest

import alchemlyb
from alchemlyb.estimators import TI_GQ, TI
from alchemlyb.parsing import amber


@pytest.fixture
def ethanol_Coulomb(gmx_ethanol_Coulomb_dHdl):
    dHdl = alchemlyb.concat(gmx_ethanol_Coulomb_dHdl)
    return dHdl


@pytest.fixture
def ethanol_VDW(gmx_ethanol_VDW_dHdl):
    dHdl = alchemlyb.concat(gmx_ethanol_VDW_dHdl)
    return dHdl


@pytest.fixture
def ethanol(ethanol_Coulomb, ethanol_VDW):
    dHdl = alchemlyb.concat([ethanol_Coulomb, ethanol_VDW])
    return dHdl


@pytest.fixture
def tyk2_complex(amber_tyk2_example_complex):
    dHdl = alchemlyb.concat(amber_tyk2_example_complex)
    return dHdl


@pytest.fixture
def benzene_VDW(gmx_benzene_VDW_dHdl):
    dHdl = alchemlyb.concat(gmx_benzene_VDW_dHdl)
    return dHdl


@pytest.fixture
def ethanol_lambdas_means_variances_index(ethanol):
    dHdl = ethanol.sort_index(level=ethanol.index.names[1:])
    means = dHdl.groupby(level=dHdl.index.names[1:]).mean()
    variances = np.square(dHdl.groupby(level=dHdl.index.names[1:]).sem())
    lambdas, new_means, new_variances, index_list = TI_GQ().separate_mean_variance(
        means, variances
    )
    return lambdas, new_means, new_variances, index_list


@pytest.fixture
def tyk2_complex_lambdas_means_variances_index(tyk2_complex):
    dHdl = tyk2_complex.sort_index(level=tyk2_complex.index.names[1:])
    means = dHdl.groupby(level=dHdl.index.names[1:]).mean()
    variances = np.square(dHdl.groupby(level=dHdl.index.names[1:]).sem())
    lambdas, new_means, new_variances, index_list = TI_GQ().separate_mean_variance(
        means, variances
    )
    return lambdas, new_means, new_variances, index_list


class TIestimatorMixin:
    def test_get_delta_f(self, X_delta_f):
        dHdl, E, dE = X_delta_f
        est = self.cls().fit(dHdl)
        # Use .iloc[0, -1] as we want to cater for both
        # delta_f_.loc[0.0, 1.0] and delta_f_.loc[(0.0, 0.0), (0.0, 1.0)]
        delta_f = est.delta_f_.iloc[0, -1]
        d_delta_f = est.d_delta_f_.iloc[0, -1]

        assert E == pytest.approx(delta_f, rel=1e-3)
        assert dE == pytest.approx(d_delta_f, rel=1e-3)


class TestTIGQ(TIestimatorMixin):
    """Tests for TI_GQ."""

    cls = TI_GQ

    T = 300.0
    kT_amber = amber.k_b * T

    @pytest.fixture(
        params=[
            ("ethanol_Coulomb", 10.565, 0.03002),
            ("ethanol_VDW", -3.386, 0.05707),
            ("ethanol", 7.179, 0.06449),
            ("tyk2_complex", -50.504, 0.09930),
        ],
    )
    def X_delta_f(self, request):
        get_dHdl, E, dE = request.param
        return request.getfixturevalue(get_dHdl), E, dE


class Test_TI_GQ_separate_mean_and_variance_multi_column:
    """Tests for TI_GQ separate_mean_and_variance function with multiple-column data"""

    def test_lambda_list(self, ethanol_lambdas_means_variances_index):
        assert isinstance(ethanol_lambdas_means_variances_index[0], list)

    def test_mean_list(self, ethanol_lambdas_means_variances_index):
        assert all(
            [
                isinstance(means, pd.Series)
                for means in ethanol_lambdas_means_variances_index[1]
            ]
        )

    def test_variance_list(self, ethanol_lambdas_means_variances_index):
        assert all(
            [
                isinstance(variances, pd.Series)
                for variances in ethanol_lambdas_means_variances_index[2]
            ]
        )

    def test_data_length(self, ethanol_lambdas_means_variances_index):
        assert sorted(
            [len(means) for means in ethanol_lambdas_means_variances_index[1]]
        ) == [
            12,
            12,
        ]

    def test_index_length(self, ethanol_lambdas_means_variances_index):
        assert len(ethanol_lambdas_means_variances_index[3]) == 26


class Test_TI_GQ_separate_mean_and_variance_single_column:
    """Tests for TI_GQ separate_mean_and_variance function with single-column data"""

    def test_lambda_list(self, tyk2_complex_lambdas_means_variances_index):
        assert isinstance(tyk2_complex_lambdas_means_variances_index[0], list)

    def test_mean_list(self, tyk2_complex_lambdas_means_variances_index):
        assert all(
            [
                isinstance(means, pd.Series)
                for means in tyk2_complex_lambdas_means_variances_index[1]
            ]
        )

    def test_variance_list(self, tyk2_complex_lambdas_means_variances_index):
        assert all(
            [
                isinstance(variances, pd.Series)
                for variances in tyk2_complex_lambdas_means_variances_index[2]
            ]
        )

    def test_data_length(self, tyk2_complex_lambdas_means_variances_index):
        assert [
            len(means) for means in tyk2_complex_lambdas_means_variances_index[1]
        ] == [12]

    def test_index_length(self, tyk2_complex_lambdas_means_variances_index):
        assert len(tyk2_complex_lambdas_means_variances_index[3]) == 14


def test_TI_GQ_separate_mean_variance_no_pertubed(tyk2_complex):
    """The test for the case where two lambda are there and one is not pertubed"""
    dHdl = tyk2_complex
    dHdl.insert(1, "bound-lambda", [1.0] * len(dHdl))
    dHdl.insert(1, "bound", [1.0] * len(dHdl))
    dHdl.set_index("bound-lambda", append=True, inplace=True)
    means = dHdl.groupby(level=dHdl.index.names[1:]).mean()
    variances = np.square(dHdl.groupby(level=dHdl.index.names[1:]).sem())
    _, new_means, _, _ = TI_GQ().separate_mean_variance(means, variances)
    assert all([isinstance(dhdl, pd.Series) for dhdl in new_means])
    assert [len(dhdl) for dhdl in new_means] == [12]


def test_TI_GQ_not_quadrature_points(benzene_VDW):
    """The test for the case where the simulation lambdas are not quadrature points"""
    dHdl = benzene_VDW
    with pytest.raises(ValueError):
        TI_GQ().fit(dHdl)


def test_TI_GQ_unsupported_lambda_numbers(tyk2_complex):
    """The test for the case where there are more lambdas than supported"""
    dHdl_1 = tyk2_complex
    # add a second copy and change it lambda values
    dHdl_2 = copy.deepcopy(dHdl_1)
    dHdl_2.reset_index(inplace=True)
    dHdl_2.lambdas += 0.01
    dHdl_2.set_index(["time", "lambdas"], inplace=True)
    # combine the to copys to have a dataset with more lambdas than supported
    dHdl = alchemlyb.concat([dHdl_1, dHdl_2])
    with pytest.raises(ValueError):
        TI_GQ().fit(dHdl)


def test_TI_GQ_multi_lambda_scaling(ethanol_Coulomb):
    """The test for the case where multiple lambdas are scaled simultaneously"""
    dHdl = ethanol_Coulomb
    # change the second lambda from all zeros to the same as the first lambda
    dHdl.reset_index(inplace=True)
    dHdl["vdw-lambda"] = dHdl["coul-lambda"]
    dHdl.set_index(["time", "coul-lambda", "vdw-lambda"], inplace=True)
    TI_GQ().fit(dHdl)


class Test_Units:
    """Test the units."""

    def test_ti(self, tyk2_complex):
        ti = TI_GQ().fit(tyk2_complex)
        assert ti.delta_f_.attrs["temperature"] == 300
        assert ti.delta_f_.attrs["energy_unit"] == "kT"
        assert ti.d_delta_f_.attrs["temperature"] == 300
        assert ti.d_delta_f_.attrs["energy_unit"] == "kT"
        assert ti.dhdl.attrs["temperature"] == 300
        assert ti.dhdl.attrs["energy_unit"] == "kT"

    def test_ti_separate_mean_variance(
        self, tyk2_complex_lambdas_means_variances_index
    ):
        dhdl_list = tyk2_complex_lambdas_means_variances_index[1]
        for dhdl in dhdl_list:
            assert dhdl.attrs["temperature"] == 300
            assert dhdl.attrs["energy_unit"] == "kT"


class Test_MultipleColumnUnits:
    """Test the case where the index has multiple columns"""

    @staticmethod
    @pytest.fixture
    def dhdl(ethanol):
        dhdl = ethanol
        return dhdl

    def test_ti_separate_mean_variance(self, ethanol_lambdas_means_variances_index):
        dhdl_list = ethanol_lambdas_means_variances_index[1]
        for dhdl in dhdl_list:
            assert dhdl.attrs["temperature"] == 300
            assert dhdl.attrs["energy_unit"] == "kT"


def test_TI_TIGQ_comparision(ethanol):
    """Test for comparing TI and TI_GQ results"""

    ti = TI().fit(ethanol)
    ti_gq = TI_GQ().fit(ethanol)
    ti_energy_results = ti.delta_f_.iloc[0, -1]
    ti_gq_energy_results = ti_gq.delta_f_.iloc[0, -1]
    ti_variance_results = ti.d_delta_f_.iloc[0, -1]
    ti_gq_variance_results = ti_gq.d_delta_f_.iloc[0, -1]

    assert ti_energy_results == pytest.approx(ti_gq_energy_results, rel=1e-1)
    assert ti_variance_results == pytest.approx(ti_gq_variance_results, rel=1e-1)



import pandas as pd
from loguru import logger

from . import _init_attrs


def _read_parquet_with_metadata(path: str, T: float) -> pd.DataFrame:
    """
    Check if the metadata is included in the Dataframe and has the correct
    temperature.

    Parameters
    ----------
    path : str
        Path to parquet file to extract dataframe from.
    T : float
        Temperature in Kelvin of the simulations.

    Returns
    -------
    DataFrame
    """
    df = pd.read_parquet(path)
    if "temperature" not in df.attrs:
        logger.warning(
            f"No temperature metadata found in {path}. "
            f"Serialise the Dataframe with pandas>=2.1 to preserve the metadata."
        )
        df.attrs["temperature"] = T
        df.attrs["energy_unit"] = "kT"
    else:
        if df.attrs["temperature"] != T:
            raise ValueError(
                f"Temperature in the input ({T}) doesn't match the temperature "
                f"in the dataframe ({df.attrs['temperature']})."
            )
    return df


def extract_u_nk(path: str, T: float) -> pd.DataFrame:
    r"""Return reduced potentials `u_nk` (unit: kT) from a pandas parquet file.

    The parquet file should be serialised from the dataframe output
    from any parser with command
    (``u_nk_df.to_parquet(path=path, index=True)``).

    Parameters
    ----------
    path : str
        Path to parquet file to extract dataframe from.
    T : float
        Temperature in Kelvin of the simulations.

    Returns
    -------
    u_nk : DataFrame
        Potential energy for each alchemical state (k) for each frame (n).


    Note
    ----
    pyarraw serializers would handle the float or string column name fine but will
    convert multi-lambda column name from `(0.0, 0.0)` to `"('0.0', '0.0')"`.
    This parser will restore the correct column name.
    Also parquet serialisation doesn't preserve the :attr:`pandas.DataFrame.attrs`.
    So the temperature is assigned in this function.


    .. versionadded:: 2.1.0

    """
    u_nk = _read_parquet_with_metadata(path, T)
    columns = list(u_nk.columns)
    if isinstance(columns[0], str) and columns[0][0] == "(":
        new_columns = []
        for column in columns:
            new_columns.append(
                tuple(
                    map(
                        float, column[1:-1].replace('"', "").replace("'", "").split(",")
                    )
                )
            )
        u_nk.columns = new_columns
    return u_nk


@_init_attrs
def extract_dHdl(path: str, T: float) -> pd.DataFrame:
    r"""Return gradients `dH/dl` (unit: kT) from a pandas parquet file.

    The parquet file should be serialised from the dataframe output
    from any parser with command
    (`dHdl_df.to_parquet(path=path, index=True)`).

    Parameters
    ----------
    path : str
        Path to parquet file to extract dataframe from.
    T : float
        Temperature in Kelvin the simulations sampled.

    Returns
    -------
    dH/dl : DataFrame
        dH/dl as a function of time for this lambda window.

    Note
    ----
    Parquet serialisation doesn't preserve the :attr:`pandas.DataFrame.attrs`.
    So the temperature is assigned in this function.


    .. versionadded:: 2.1.0

    """
    return _read_parquet_with_metadata(path, T)



"""Collection of utilities used by many parsers."""

import bz2
import gzip
import os

from os import PathLike
from typing import TextIO


def bz2_open(filename: str, mode: str) -> TextIO:
    mode += "t" if mode in ["r", "w", "a", "x"] else ""
    return bz2.open(filename, mode)  # type: ignore[return-value]


def gzip_open(filename: str, mode: str) -> TextIO:
    mode += "t" if mode in ["r", "w", "a", "x"] else ""
    return gzip.open(filename, mode)  # type: ignore[return-value]


def anyopen(
    datafile: PathLike | TextIO, mode: str = "r", compression: None | str = None
) -> TextIO:
    """Return a file stream for file or stream, even if compressed.

    Supports files compressed with bzip2 (.bz2) and gzip (.gz) compression
    schemes. The appropriate extension must be present for the function to
    properly handle the file without specifying `compression`.

    If giving a stream for `datafile`, then you must specify `compression` if
    the stream is compressed. Otherwise the stream will be passed through
    as-is.

    If `datafile` is a filepath, then `compression` will take precedence over
    any extension on the filename. Leaving `compression` as `None` will rely on
    the extension for determining compression, if any.

    .. versionchanged:: 0.7.0
       Removed stated support for zip, given broken implementation.

    Parameters
    ----------
    datafile : PathLike | IO
        Path to file to use, or an open IO stream. If an IO stream, use
        `compression` to specify the type of compression used, if any.
    mode : str
        Mode for stream; usually 'r' or 'w'.
    compression : str
        Use to specify compression. Must be one of 'bzip2', 'gzip'.
        Overrides use of extension for determining compression if `datafile` is
        a file.

        .. versionadded:: 0.7.0

    Returns
    -------
    stream : stream
        Open stream for reading or writing, depending on mode.

        .. versionchanged:: 0.7.0
           Explicit support for writing added.

    """
    # opener for each type of file
    extensions = {".bz2": bz2_open, ".gz": gzip_open}

    # compression selections available
    compressions = {"bzip2": bz2_open, "gzip": gzip_open}

    # if `datafile` is a stream
    if (hasattr(datafile, "read") and any((i in mode for i in ("r",)))) or (
        hasattr(datafile, "write") and any((i in mode for i in ("w", "a", "x")))
    ):
        # if no compression specified, just pass the stream through
        if compression is None:
            return datafile  # type: ignore[return-value]
        elif compression in compressions:
            compressor = compressions[compression]
            return compressor(datafile, mode=mode)  # type: ignore[arg-type]
        else:
            raise ValueError(
                "`datafile` is a stream"
                + ", but specified `compression` '{compression}' is not supported"
            )

    # otherwise, treat as a file
    # allow compression to override any extension on the file
    if compression in compressions:
        opener = compressions[compression]

    # use extension to determine the compression used, if present
    elif compression is None:
        ext = os.path.splitext(datafile)[1]  # type: ignore[arg-type]
        if ext in extensions:
            opener = extensions[ext]
        else:
            opener = open  # type: ignore[assignment]

    return opener(datafile, mode)  # type: ignore[arg-type]



from functools import wraps
from typing import Any, Callable
import pandas as pd


def _init_attrs(
    func: Callable[..., None | pd.DataFrame],
) -> Callable[..., None | pd.DataFrame]:
    """Add temperature to the parsed dataframe.

    The temperature is added to the dataframe as dataframe.attrs['temperature']
    and the energy unit is initiated as dataframe.attrs['energy_unit'] = 'kT'.
    """

    @wraps(func)
    def wrapper(
        outfile: str, T: float, *args: Any, **kwargs: Any
    ) -> None | pd.DataFrame:
        dataframe = func(outfile, T, *args, **kwargs)
        if dataframe is not None:
            dataframe.attrs["temperature"] = T
            dataframe.attrs["energy_unit"] = "kT"
        return dataframe

    return wrapper


def _init_attrs_dict(
    func: Callable[..., dict[str, None | pd.DataFrame]],
) -> Callable[..., dict[str, None | pd.DataFrame]]:
    """Add temperature and energy units to the parsed dataframes.

    The temperature is added to the dataframe as dataframe.attrs['temperature']
    and the energy unit is initiated as dataframe.attrs['energy_unit'] = 'kT'.
    """

    @wraps(func)
    def wrapper(
        outfile: str, T: float, *args: Any, **kwargs: Any
    ) -> dict[str, None | pd.DataFrame]:
        dict_with_df = func(outfile, T, *args, **kwargs)
        for k in dict_with_df.keys():
            dataframe = dict_with_df[k]
            if dataframe is not None:
                dataframe.attrs["temperature"] = T
                dataframe.attrs["energy_unit"] = "kT"
        return dict_with_df

    return wrapper



"""
The :mod:`alchemlyb.preprocessing` module includes methods for subsampling and
preparing data for estimators.
"""

from .subsampling import equilibrium_detection
from .subsampling import (
    slicing,
    dhdl2series,
    u_nk2series,
    decorrelate_dhdl,
    decorrelate_u_nk,
)
from .subsampling import statistical_inefficiency

__all__ = [
    "slicing",
    "statistical_inefficiency",
    "equilibrium_detection",
    "decorrelate_dhdl",
    "decorrelate_u_nk",
    "dhdl2series",
    "u_nk2series",
]



"""Parsers for extracting alchemical data from `AMBER <http://ambermd.org>`_ output files.

Some of the file parsing parts are adapted from
`alchemical-analysis`_.

.. _alchemical-analysis: https://github.com/MobleyLab/alchemical-analysis

.. versionchanged:: 1.0.0
    Now raises :exc:`ValueError` when an invalid file is given to the parser.
    Now raises :exc:`ValueError` when inconsistency in MBAR states/data is found.

"""

import re
from typing import Any, Iterator

import numpy as np
import pandas as pd
from loguru import logger

from . import _init_attrs_dict
from .util import anyopen
from ..postprocessors.units import R_kJmol, kJ2kcal

k_b = R_kJmol * kJ2kcal

_FP_RE = r"[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?"


def convert_to_pandas(file_datum: Any) -> pd.DataFrame:
    """Convert the data structure from numpy to pandas format"""
    data_dic: dict[str, list[Any]] = {}
    data_dic["dHdl"] = []
    data_dic["lambdas"] = []
    data_dic["time"] = []
    for frame_index, frame_dhdl in enumerate(file_datum.gradients):
        data_dic["dHdl"].append(frame_dhdl)
        data_dic["lambdas"].append(file_datum.clambda)
        frame_time = file_datum.t0 + (frame_index + 1) * file_datum.dt * file_datum.ntpr
        data_dic["time"].append(frame_time)
    df = pd.DataFrame(
        data_dic["dHdl"],
        columns=["dHdl"],
        index=pd.Index(data_dic["time"], name="time", dtype="Float64"),
    )
    df["lambdas"] = data_dic["lambdas"][0]
    df = df.reset_index().set_index(["time"] + ["lambdas"])
    return df


def _pre_gen(it: Iterator[Any], first: None | Any) -> Iterator[Any]:
    """A generator that returns first first if it exists."""

    if first:
        yield first

    while it:
        try:
            yield next(it)
        except StopIteration:
            return


class SectionParser:
    """
    A simple parser to extract data values from sections.
    """

    def __init__(self, filename: str) -> None:
        """Opens a file according to its file type."""
        self.filename = filename
        try:
            self.fileh = anyopen(self.filename, "r")  # type: ignore[arg-type]
        except Exception:
            logger.exception("Cannot open file {}", filename)
            raise
        self.lineno = 0

    def skip_lines(self, nlines: int) -> None | str:
        """Skip a given number of lines."""
        lineno = 0
        for line in self:
            lineno += 1
            if lineno > nlines:
                return line
        return None

    def skip_after(self, pattern: str) -> bool:
        """Skip until after a line that matches a regex pattern."""
        Found_pattern = False
        for line in self:
            match = re.search(pattern, line)
            if match:
                Found_pattern = True
                break
        return Found_pattern

    def extract_section(
        self,
        start: str,
        end: str,
        fields: list[str],
        limit: None | str = None,
        extra: str = "",
    ) -> list[float]:
        """
        Extract data values (int, float) in fields from a section
        marked with start and end regexes.  Do not read further than
        limit regex.
        """
        inside = False
        lines = []
        for line in _pre_gen(self, extra):
            if limit and re.search(limit, line):
                break
            if re.search(start, line):
                inside = True
            if inside:
                if re.search(end, line):
                    break
                lines.append(line.rstrip("\n"))
        line = "".join(lines)
        result = []
        for field in fields:
            match = re.search(rf" {field}\s*=\s*(\*+|{_FP_RE}|\d+)", line)
            if match:
                value = match.group(1)
                if "*" in value:  # catch fortran format overflow
                    result.append(float("Inf"))
                else:
                    try:
                        result.append(int(value))
                    except ValueError:
                        result.append(float(value))
            else:  # section may be incomplete
                result.append(None)  # type: ignore[arg-type]
        return result

    def __iter__(self) -> Iterator[str]:
        return self

    def __next__(self) -> str:
        """Read next line of the filehandle and check for EOF."""
        self.lineno += 1
        return next(self.fileh)

    def close(self) -> None:
        """Close the filehandle."""
        self.fileh.close()

    def __enter__(self) -> "SectionParser":
        return self

    def __exit__(self, typ: Any, value: Any, traceback: Any) -> None:
        self.close()


class FEData:
    """A simple struct container to collect data from individual files."""

    __slots__ = [
        "clambda",
        "t0",
        "dt",
        "T",
        "ntpr",
        "bar_intervall",
        "gradients",
        "mbar_energies",
        "have_mbar",
        "mbar_lambdas",
        "mbar_lambda_idx",
    ]

    def __init__(self) -> None:
        self.clambda = -1.0
        self.t0 = -1.0
        self.dt = -1.0
        self.T = -1.0
        self.ntpr = -1
        self.bar_intervall = -1
        self.gradients: list[float] = []
        self.mbar_energies: list[list[float]] = []
        self.have_mbar = False
        self.mbar_lambdas: list[str] = []
        self.mbar_lambda_idx = -1


def file_validation(outfile: str) -> FEData:
    """
    Function that validate and parse an AMBER output file.
    :exc:`ValueError` are risen if inconsinstencies in the input file are found.

    Parameters
    ----------
    outfile : str
        Path to AMBER .out file to validate and extract data from.

    Returns
    -------
    `:class:~FEData`
        FEData object populated with data from the parsed AMBER output file.

    """

    file_datum = FEData()
    with SectionParser(outfile) as secp:
        line = secp.skip_lines(5)

        if not line:
            logger.error("The file {} does not contain any data, it's empty.", outfile)
            raise ValueError(f"file {outfile} does not contain any data.")

        if not secp.skip_after("^   2.  CONTROL  DATA  FOR  THE  RUN"):
            logger.error('No "CONTROL DATA" section found in file {}.', outfile)
            raise ValueError(f'no "CONTROL DATA" section found in file {outfile}')

        (ntpr,) = secp.extract_section("^Nature and format of output:", "^$", ["ntpr"])
        nstlim, dt = secp.extract_section("Molecular dynamics:", "^$", ["nstlim", "dt"])
        (T,) = secp.extract_section("temperature regulation:", "^$", ["temp0"])
        if not T:
            logger.error('No valid "temp0" record found in file {}.', outfile)
            raise ValueError(f'no valid "temp0" record found in file {outfile}')

        (clambda,) = secp.extract_section(
            "^Free energy options:", "^$", ["clambda"], "^---"
        )
        if clambda is None:
            logger.error(  # type: ignore[unreachable]
                'No free energy section found in file {}, "clambda" was None.', outfile
            )
            raise ValueError(f"no free energy section found in file {outfile}")

        have_mbar, bar_intervall, mbar_states = secp.extract_section(
            "^FEP MBAR options:",
            "^$",
            ["ifmbar", "bar_intervall", "mbar_states"],
            "^---",
        )
        if have_mbar:
            mbar_lambdas = _process_mbar_lambdas(secp)
            file_datum.mbar_lambdas = mbar_lambdas
            clambda_str = f"{clambda:6.4f}"

            if clambda_str not in mbar_lambdas:
                logger.warning(
                    "WARNING: lamba {} not contained in set of "
                    "MBAR lambas: {}\nNot using MBAR.",
                    clambda_str,
                    ", ".join(mbar_lambdas),
                )
                have_mbar = False
            else:
                mbar_nlambda = len(mbar_lambdas)
                if mbar_nlambda != mbar_states:
                    logger.error(
                        "the number of lambda windows read ({})"
                        "is different from what expected ({})",
                        ",".join(mbar_lambdas),
                        mbar_states,
                    )
                    raise ValueError(
                        f"the number of lambda windows read ({mbar_nlambda})"
                        f" is different from what expected ({mbar_states})"
                    )
                mbar_lambda_idx = mbar_lambdas.index(clambda_str)
                file_datum.mbar_lambda_idx = mbar_lambda_idx

                for _ in range(mbar_nlambda):
                    file_datum.mbar_energies.append([])

        if not secp.skip_after("^   3.  ATOMIC "):
            logger.error('No "ATOMIC" section found in the file {}.', outfile)
            raise ValueError(f'no "ATOMIC" section found in file {outfile}')

        (t0,) = secp.extract_section("^ begin time", "^$", ["coords"])
        if t0 is None:
            logger.error("No starting simulation time in file {}.", outfile)  # type: ignore[unreachable]
            raise ValueError(f"No starting simulation time in file {outfile}")

        if not secp.skip_after("^   4.  RESULTS"):
            logger.error('No "RESULTS" section found in the file {}.', outfile)
            raise ValueError(f'no "RESULTS" section found in file {outfile}')

    file_datum.clambda = clambda
    file_datum.t0 = t0
    file_datum.dt = dt
    file_datum.ntpr = ntpr  # type: ignore[assignment]
    file_datum.bar_intervall = bar_intervall  # type: ignore[assignment]
    file_datum.T = T
    file_datum.have_mbar = have_mbar  # type: ignore[assignment]

    return file_datum


@_init_attrs_dict
def extract(outfile: str, T: float) -> dict[str, None | pd.DataFrame]:
    """Return reduced potentials `u_nk` and gradients `dH/dl` from AMBER outputfile.

    Parameters
    ----------
    outfile : str
        Path to AMBER .out file to extract data from.
    T : float
        Temperature in Kelvin at which the simulations were performed;
        needed to generated the reduced potential (in units of kT)

    Returns
    -------
    Dict
        A dictionary with keys of 'u_nk', which is a pandas DataFrame of reduced potentials for each
        alchemical state (k) for each frame (n), and 'dHdl', which is a Series of dH/dl
        as a function of time for this lambda window.


    .. versionadded:: 1.0.0

    """

    beta = 1 / (k_b * T)

    file_datum = file_validation(outfile)

    if not np.isclose(T, file_datum.T, atol=0.01):
        msg = f"The temperature read from the input file ({file_datum.T:.2f} K)"
        msg += f" is different from the temperature passed as parameter ({T:.2f} K)"
        logger.error(msg)
        raise ValueError(msg)

    finished = False
    with SectionParser(outfile) as secp:
        line = secp.skip_lines(5)
        high_E_cnt = 0
        nensec = 0
        old_nstep = -1
        for line in secp:
            if "      A V E R A G E S   O V E R" in line:
                _ = secp.skip_after("^|=========================================")
            elif line.startswith(" NSTEP"):
                nstep, dvdl = secp.extract_section(
                    "^ NSTEP", "^ ---", ["NSTEP", "DV/DL"], extra=line
                )
                if nstep != old_nstep and dvdl is not None and nstep is not None:
                    if finished:
                        raise ValueError(
                            "TI Energy detected after the TIMINGS section."
                            + " Did you concatenate the output file?"
                        )
                    file_datum.gradients.append(dvdl)
                    nensec += 1
                    old_nstep = nstep  # type: ignore[assignment]
            elif line.startswith("MBAR Energy analysis") and file_datum.have_mbar:
                if finished:
                    raise ValueError(
                        "MBAR Energy detected after the TIMINGS section."
                        + " Did you concatenate the output file?"
                    )
                mbar = secp.extract_section(
                    "^MBAR", "^ ---", file_datum.mbar_lambdas, extra=line
                )

                if None in mbar:
                    msg = "Something strange parsing the following MBAR section."
                    msg += "\nMaybe the mbar_lambda values are incorrect?"
                    logger.error("{}\n{}", msg, mbar)
                    raise ValueError(msg)

                reference_energy = mbar[file_datum.mbar_lambda_idx]
                for lmbda, energy in enumerate(mbar):
                    if energy > 0.0:
                        high_E_cnt += 1

                    file_datum.mbar_energies[lmbda].append(
                        beta * (energy - reference_energy)
                    )
            elif line == "   5.  TIMINGS\n":
                finished = True

        if high_E_cnt:
            logger.warning(
                "{} MBAR energ{} > 0.0 kcal/mol",
                high_E_cnt,
                "ies are" if high_E_cnt > 1 else "y is",
            )

    if not finished:
        logger.warning("WARNING: file {} is a prematurely terminated run", outfile)

    if file_datum.have_mbar:
        mbar_time = [
            file_datum.t0 + (frame_index + 1) * file_datum.dt * file_datum.bar_intervall
            for frame_index in range(len(file_datum.mbar_energies[0]))
        ]

        mbar_df = pd.DataFrame(
            file_datum.mbar_energies,
            index=np.array(file_datum.mbar_lambdas, dtype=np.float64),
            columns=pd.MultiIndex.from_arrays(
                [mbar_time, np.repeat(file_datum.clambda, len(mbar_time))],
                names=["time", "lambdas"],
            ),
        ).T
    else:
        logger.info('WARNING: No MBAR energies found! "u_nk" entry will be None')
        mbar_df = None

    if not nensec:
        logger.warning("WARNING: File {} does not contain any dV/dl data", outfile)
        dHdl_df = None
    else:
        logger.info("Read {} dV/dl data points in file {}", nensec, outfile)
        dHdl_df = convert_to_pandas(file_datum)
        dHdl_df["dHdl"] *= beta

    return {"u_nk": mbar_df, "dHdl": dHdl_df}


def extract_dHdl(outfile: str, T: float) -> None | pd.DataFrame:
    """Return gradients `dH/dl` from AMBER TI outputfile.

    Parameters
    ----------
    outfile : str
        Path to AMBER .out file to extract data from.
    T : float
        Temperature in Kelvin at which the simulations were performed

    Returns
    -------
    dH/dl : Series
        dH/dl as a function of time for this lambda window.


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.

    """
    extracted = extract(outfile, T)
    return extracted["dHdl"]


def extract_u_nk(outfile: str, T: float) -> None | pd.DataFrame:
    """Return reduced potentials `u_nk` from AMBER outputfile.

    Parameters
    ----------
    outfile : str
        Path to AMBER .out file to extract data from.
    T : float
        Temperature in Kelvin at which the simulations were performed;
        needed to generated the reduced potential (in units of kT)

    Returns
    -------
    u_nk : DataFrame
        Reduced potential for each alchemical state (k) for each frame (n).


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.

    """
    extracted = extract(outfile, T)
    return extracted["u_nk"]


def _process_mbar_lambdas(secp: SectionParser) -> list[str]:
    """
    Extract the lambda points used to compute MBAR energies from an AMBER MDOUT file.
    Parameters
    ----------
    secp: .out file from amber simulation.

    Returns
    -------
    mbar_lambdas: lambda values used for MBAR energy collection in simulation.

    """

    in_mbar = False
    mbar_lambdas = []

    for line in secp:
        if line.startswith("    MBAR - lambda values considered:"):
            in_mbar = True
            continue

        if in_mbar:
            if line.startswith("    Extra"):
                break

            if "total" in line:
                data = line.split()
                mbar_lambdas.extend(data[2:])
            else:
                # AMBER splits the MBAR states in more than one line
                # if there are many of them
                mbar_lambdas.extend(line.split())

    return mbar_lambdas



"""Tests for preprocessing functions."""

import logging

import numpy as np
import pytest
from numpy.testing import assert_allclose

import alchemlyb
from alchemlyb.preprocessing import (
    slicing,
    statistical_inefficiency,
    equilibrium_detection,
    decorrelate_u_nk,
    decorrelate_dhdl,
    u_nk2series,
    dhdl2series,
)


def _check_data_is_outside_bounds(data, lower, upper):
    """
    Helper function to make sure that `data` has entries that are
    below the `lower` bound, and above the `upper` bound.
    This is used by slicing tests to make sure that the data
    provided is appropriate for the tests.
    """
    assert any(data.reset_index()["time"] < lower)
    assert any(data.reset_index()["time"] > upper)


@pytest.fixture()
def dHdl(gmx_benzene_Coulomb_dHdl):
    return gmx_benzene_Coulomb_dHdl[0]


@pytest.fixture()
def u_nk(gmx_benzene_Coulomb_u_nk):
    return gmx_benzene_Coulomb_u_nk[0]


@pytest.fixture()
def multi_index_u_nk(gmx_ABFE_complex_u_nk):
    return gmx_ABFE_complex_u_nk[0]


@pytest.fixture()
def multi_index_dHdl(gmx_ABFE_complex_dHdl):
    return gmx_ABFE_complex_dHdl[0]


class TestSlicing:
    """Test slicing functionality."""

    def slicer(self, *args, **kwargs):
        return slicing(*args, **kwargs)

    @pytest.mark.parametrize(("data", "size"), [("dHdl", 661), ("u_nk", 661)])
    def test_basic_slicing(self, data, size, request):
        assert (
            len(
                self.slicer(
                    request.getfixturevalue(data), lower=1000, upper=34000, step=5
                )
            )
            == size
        )

    def test_unchanged(self, namd_idws):
        # NAMD energy files only have dE for adjacent lambdas, this ensures
        # that the slicer will not drop these rows as they have NaN values.
        # Do the pre-processing as the u_nk are from all lambdas
        groups = namd_idws.groupby("fep-lambda")
        for key, group in groups:
            group = group[~group.index.duplicated(keep="first")]
            df = self.slicer(group, None, None, None)
            assert len(df) == len(group)

    @pytest.mark.parametrize(
        ("dataloader", "lower", "upper"),
        [
            ("dHdl", 1000, 34000),
            ("u_nk", 1000, 34000),
        ],
    )
    def test_data_is_unchanged(self, dataloader, lower, upper, request):
        """
        Test that slicing does not change the underlying data
        """
        # Load data
        data = request.getfixturevalue(dataloader)
        # Check that the input data is appropriate for the test
        _check_data_is_outside_bounds(data, lower, upper)

        # Slice data, and test that we didn't change the input data
        original_length = len(data)
        self.slicer(data, lower=lower, upper=upper, step=5)
        assert len(data) == original_length

    @pytest.mark.parametrize(
        ("dataloader", "lower", "upper"),
        [
            ("dHdl", 1000, 34000),
            ("u_nk", 1000, 34000),
        ],
    )
    def test_lower_and_upper_bound(self, dataloader, lower, upper, request):
        """
        Test that the lower and upper time is respected
        """
        # Load data
        data = request.getfixturevalue(dataloader)
        # Check that the input data is appropriate for the test
        _check_data_is_outside_bounds(data, lower, upper)

        # Slice data, and test that we don't observe times outside
        # the prescribed range
        sliced = self.slicer(data, lower=lower, upper=upper, step=5)
        assert all(sliced.reset_index()["time"] >= lower)
        assert all(sliced.reset_index()["time"] <= upper)

    @pytest.mark.parametrize("dataloader", ["dHdl", "u_nk"])
    def test_disordered_exception(self, dataloader, request):
        """Test that a shuffled DataFrame yields a KeyError."""
        data = request.getfixturevalue(dataloader)
        indices = data.index.values
        np.random.shuffle(indices)

        df = data.loc[indices]

        with pytest.raises(KeyError):
            self.slicer(df, lower=200)

    @pytest.mark.parametrize(
        "dataloader", ["gmx_benzene_Coulomb_dHdl", "gmx_benzene_Coulomb_u_nk"]
    )
    def test_duplicated_exception(self, dataloader, request):
        """Test that a DataFrame with duplicate times yields a KeyError."""
        data = alchemlyb.concat(request.getfixturevalue(dataloader))
        with pytest.raises(KeyError):
            self.slicer(data.sort_index(axis=0), lower=200)

    def test_subsample_bounds_and_step(self, multi_index_u_nk):
        """Make sure that slicing the series also works"""
        subsample = statistical_inefficiency(
            multi_index_u_nk, multi_index_u_nk.sum(axis=1), lower=100, upper=400, step=2
        )
        assert len(subsample) == 76

    def test_multiindex_duplicated(self, multi_index_u_nk):
        subsample = statistical_inefficiency(
            multi_index_u_nk, multi_index_u_nk.sum(axis=1)
        )
        assert len(subsample) == 501

    def test_sort_off(self, multi_index_u_nk):
        unsorted = alchemlyb.concat([multi_index_u_nk[-500:], multi_index_u_nk[:500]])
        with pytest.raises(KeyError):
            statistical_inefficiency(unsorted, unsorted.sum(axis=1), sort=False)

    def test_sort_on(self, multi_index_u_nk):
        unsorted = alchemlyb.concat([multi_index_u_nk[-500:], multi_index_u_nk[:500]])
        subsample = statistical_inefficiency(unsorted, unsorted.sum(axis=1), sort=True)
        assert subsample.reset_index(0)["time"].is_monotonic_increasing

    def test_sort_on_noseries(self, multi_index_u_nk):
        unsorted = alchemlyb.concat([multi_index_u_nk[-500:], multi_index_u_nk[:500]])
        subsample = statistical_inefficiency(unsorted, None, sort=True)
        assert subsample.reset_index(0)["time"].is_monotonic_increasing

    def test_duplication_off(self, multi_index_u_nk):
        duplicated = alchemlyb.concat([multi_index_u_nk, multi_index_u_nk])
        with pytest.raises(KeyError):
            statistical_inefficiency(
                duplicated, duplicated.sum(axis=1), drop_duplicates=False
            )

    def test_duplication_on_dataframe(self, multi_index_u_nk):
        duplicated = alchemlyb.concat([multi_index_u_nk, multi_index_u_nk])
        subsample = statistical_inefficiency(
            duplicated, duplicated.sum(axis=1), drop_duplicates=True
        )
        assert len(subsample) < 1000

    def test_duplication_on_dataframe_noseries(self, multi_index_u_nk):
        duplicated = alchemlyb.concat([multi_index_u_nk, multi_index_u_nk])
        subsample = statistical_inefficiency(duplicated, None, drop_duplicates=True)
        assert len(subsample) == 1001

    def test_duplication_on_series(self, multi_index_u_nk):
        duplicated = alchemlyb.concat([multi_index_u_nk, multi_index_u_nk])
        subsample = statistical_inefficiency(
            duplicated.sum(axis=1), duplicated.sum(axis=1), drop_duplicates=True
        )
        assert len(subsample) < 1000

    def test_duplication_on_series_noseries(self, multi_index_u_nk):
        duplicated = alchemlyb.concat([multi_index_u_nk, multi_index_u_nk])
        subsample = statistical_inefficiency(
            duplicated.sum(axis=1), None, drop_duplicates=True
        )
        assert len(subsample) == 1001


class CorrelatedPreprocessors:
    @pytest.mark.parametrize(("dataloader", "size"), [("dHdl", 4001), ("u_nk", 4001)])
    def test_subsampling(self, dataloader, size, request):
        """Basic test for execution; resulting size of dataset sensitive to
        machine and depends on algorithm.
        """
        data = request.getfixturevalue(dataloader)
        assert len(self.slicer(data, series=data.loc[:, data.columns[0]])) <= size

    @pytest.mark.parametrize("dataloader", ["dHdl", "u_nk"])
    def test_no_series(self, dataloader, request):
        """Check that we get the same result as simple slicing with no Series."""
        data = request.getfixturevalue(dataloader)
        df_sub = self.slicer(data, lower=200, upper=5000, step=2)
        df_sliced = slicing(data, lower=200, upper=5000, step=2)

        assert np.all((df_sub == df_sliced))


class TestStatisticalInefficiency(TestSlicing, CorrelatedPreprocessors):
    def slicer(self, *args, **kwargs):
        return statistical_inefficiency(*args, **kwargs)

    @pytest.mark.parametrize(
        ("conservative", "dataloader", "size"),
        [
            (True, "dHdl", 2001),  # 0.00:  g = 1.0559445620585415
            (True, "u_nk", 2001),  # 'fep': g = 1.0560203916559594
            (False, "dHdl", 3789),
            (False, "u_nk", 3571),
        ],
    )
    def test_conservative(self, dataloader, size, conservative, request):
        data = request.getfixturevalue(dataloader)
        sliced = self.slicer(
            data, series=data.loc[:, data.columns[0]], conservative=conservative
        )
        # results can vary slightly with different machines
        # so possibly do
        # delta = 10
        # assert size - delta < len(sliced) < size + delta
        assert len(sliced) == size

    @pytest.mark.parametrize(
        "dataloader,end,step",
        [
            ("dHdl", 20, None),  # wrong length
            ("dHdl", None, -1),  # wrong time stamps (reversed)
        ],
    )
    def test_raise_ValueError_for_mismatched_data(self, dataloader, end, step, request):
        data = request.getfixturevalue(dataloader)
        with pytest.raises(ValueError):
            self.slicer(data, series=data["fep"][:end:step])

    @pytest.mark.parametrize(
        ("dataloader", "lower", "upper"),
        [
            ("dHdl", 1000, 34000),
            ("u_nk", 1000, 34000),
        ],
    )
    @pytest.mark.parametrize("use_series", [True, False])
    @pytest.mark.parametrize("conservative", [True, False])
    def test_data_is_unchanged(
        self, dataloader, use_series, lower, upper, conservative, request
    ):
        """
        Test that using statistical_inefficiency does not change the underlying data

        statistical_inefficiency is equivalent to slicing it its `series` parameter
        is not set. If the `series` parameter is set, additional inefficiency
        calculations are performed. We want to test both behaviors. The behavior
        is toggled using the `use_series` argument.
        """
        # Load data
        data = request.getfixturevalue(dataloader)
        # Check that the input data is appropriate for the test
        _check_data_is_outside_bounds(data, lower, upper)

        # Define subsampling series if required
        series = data.sum(axis=1) if use_series else None

        # Slice data, and test that we didn't change the input data
        original_length = len(data)
        self.slicer(
            data,
            series=series,
            lower=lower,
            upper=upper,
            step=5,
            conservative=conservative,
        )
        assert len(data) == original_length

    @pytest.mark.parametrize(
        ("dataloader", "lower", "upper"),
        [
            ("dHdl", 1000, 34000),
            ("u_nk", 1000, 34000),
        ],
    )
    @pytest.mark.parametrize("use_series", [True, False])
    @pytest.mark.parametrize("conservative", [True, False])
    def test_lower_and_upper_bound_slicer(
        self, dataloader, use_series, lower, upper, conservative, request
    ):
        """
        Test that the lower and upper time is respected when using statistical_inefficiency

        statistical_inefficiency is equivalent to slicing it its `series` parameter
        is not set. If the `series` parameter is set, additional inefficiency
        calculations are performed. We want to test both behaviors. The behavior
        is toggled using the `use_series` argument.
        """
        # Load data
        data = request.getfixturevalue(dataloader)
        # Check that the input data is appropriate for the test
        _check_data_is_outside_bounds(data, lower, upper)

        # Define subsampling series if required
        series = data.sum(axis=1) if use_series else None

        # Slice data, and test that we don't observe times outside
        # the prescribed range
        sliced = self.slicer(
            data,
            series=series,
            lower=lower,
            upper=upper,
            step=5,
            conservative=conservative,
        )
        assert all(sliced.reset_index()["time"] >= lower)
        assert all(sliced.reset_index()["time"] <= upper)

    @pytest.mark.parametrize(
        ("dataloader", "lower", "upper"),
        [
            ("dHdl", 1000, 34000),
            ("u_nk", 1000, 34000),
        ],
    )
    @pytest.mark.parametrize("conservative", [True, False])
    def test_slicing_inefficiency_equivalence(
        self, dataloader, lower, upper, conservative, request
    ):
        """
        Test that first slicing the data frame, then subsampling is equivalent to
        subsampling with lower / upper bounds set
        """
        # Load data
        data = request.getfixturevalue(dataloader)
        # Check that the input data is appropriate for the test
        _check_data_is_outside_bounds(data, lower, upper)

        # Slice dataframe, then subsample it based on the sum of its components
        sliced_data = slicing(data, lower=lower, upper=upper)
        subsampled_sliced_data = self.slicer(
            sliced_data, series=sliced_data.sum(axis=1), conservative=conservative
        )

        # Subsample the dataframe based on the sum of its components while
        # also specifying the slicing range
        subsampled_data = self.slicer(
            data,
            series=data.sum(axis=1),
            lower=lower,
            upper=upper,
            conservative=conservative,
        )

        assert (subsampled_sliced_data == subsampled_data).all(axis=None)


class TestEquilibriumDetection(TestSlicing, CorrelatedPreprocessors):
    def slicer(self, *args, **kwargs):
        return equilibrium_detection(*args, **kwargs)


class Test_Units:
    """Test the preprocessing module."""

    def test_slicing(self, u_nk):
        """Test if extract_u_nk assign the attr correctly"""
        new_u_nk = slicing(u_nk)
        assert new_u_nk.attrs["temperature"] == 300
        assert new_u_nk.attrs["energy_unit"] == "kT"

    def test_statistical_inefficiency(self, dHdl):
        """Test if extract_u_nk assign the attr correctly"""
        new_dhdl = statistical_inefficiency(dHdl)
        assert new_dhdl.attrs["temperature"] == 300
        assert new_dhdl.attrs["energy_unit"] == "kT"

    def test_equilibrium_detection(self, dHdl):
        """Test if extract_u_nk assign the attr correctly"""
        new_dhdl = equilibrium_detection(dHdl)
        assert new_dhdl.attrs["temperature"] == 300
        assert new_dhdl.attrs["energy_unit"] == "kT"


@pytest.mark.parametrize(("method", "size"), [("all", 2001), ("dE", 2001)])
def test_decorrelate_u_nk_single_l(u_nk, method, size):
    assert (
        len(decorrelate_u_nk(u_nk, method=method, drop_duplicates=True, sort=True))
        == size
    )


def test_decorrelate_u_nk_burnin(u_nk):
    assert (
        len(
            decorrelate_u_nk(
                u_nk,
                method="dE",
                drop_duplicates=True,
                sort=True,
                remove_burnin=True,
            )
        )
        == 2848
    )


def test_decorrelate_dhdl_burnin(dHdl):
    assert (
        len(
            decorrelate_dhdl(
                dHdl,
                drop_duplicates=True,
                sort=True,
                remove_burnin=True,
            )
        )
        == 2848
    )


@pytest.mark.parametrize(("method", "size"), [("all", 501), ("dE", 501)])
def test_decorrelate_u_nk_multiple_l(multi_index_u_nk, method, size):
    assert (
        len(
            decorrelate_u_nk(
                multi_index_u_nk,
                method=method,
            )
        )
        == size
    )


def test_decorrelate_dhdl_single_l(u_nk):
    assert len(decorrelate_dhdl(u_nk, drop_duplicates=True, sort=True)) == 2001


def test_decorrelate_dhdl_multiple_l(multi_index_dHdl):
    assert (
        len(
            decorrelate_dhdl(
                multi_index_dHdl,
            )
        )
        == 501
    )


def test_raise_nou_nk(multi_index_dHdl):
    with pytest.raises(ValueError):
        decorrelate_u_nk(
            multi_index_dHdl,
        )


class TestDhdl2series:
    @pytest.mark.parametrize("methodargs", [{}, {"method": "all"}])
    def test_dhdl2series(self, dHdl, methodargs):
        series = dhdl2series(dHdl, **methodargs)
        assert len(series) == len(dHdl)
        assert_allclose(series, dHdl.sum(axis=1))

    def test_other_method_ValueError(self, dHdl):
        with pytest.raises(
            ValueError, match="Only method='all' is supported for dhdl2series()."
        ):
            dhdl2series(dHdl, method="dE")


class TestU_nk2series:
    @pytest.mark.parametrize(
        "methodargs,reference",  # reference = sum
        [
            ({}, 7988.667045),
            ({"method": "all"}, 85982.34668751864),
            ({"method": "dE"}, 7988.667045),
        ],
    )
    def test_u_nk2series(self, u_nk, methodargs, reference):
        series = u_nk2series(u_nk, **methodargs)
        assert len(series) == len(u_nk)
        assert_allclose(series.sum(), reference)

    @pytest.mark.parametrize(
        "methodargs,reference",  # reference = sum
        [
            ({"method": "dhdl_all"}, 85982.34668751864),
            ({"method": "dhdl"}, 7988.667045),
        ],
    )
    def test_u_nk2series_deprecated(self, u_nk, methodargs, reference):
        with pytest.warns(
            DeprecationWarning,
            match=r"Method 'dhdl.*' has been deprecated, using '.*' instead\. "
            r"'dhdl.*' will be removed in alchemlyb 3\.0\.0\.",
        ):
            series = u_nk2series(u_nk, **methodargs)
        assert len(series) == len(u_nk)
        assert_allclose(series.sum(), reference)

    def test_other_method_ValueError(self, u_nk):
        with pytest.raises(ValueError, match="Decorrelation method bogus not found."):
            u_nk2series(u_nk, method="bogus")


class TestLogging:
    def test_detect_equilibration(self, caplog, u_nk):
        with caplog.at_level(logging.DEBUG):
            decorrelate_u_nk(u_nk, remove_burnin=True)

            assert "Running equilibration detection." in caplog.text
            assert "Start index:" in caplog.text
            assert "Statistical inefficiency:" in caplog.text
            assert "Number of uncorrelated samples:" in caplog.text

    def test_statistical_inefficiency(self, caplog, u_nk):
        with caplog.at_level(logging.DEBUG):
            decorrelate_u_nk(u_nk)

            assert "Running statistical inefficiency analysis." in caplog.text
            assert "Statistical inefficiency:" in caplog.text
            assert "Number of uncorrelated samples:" in caplog.text


def test_unequil_input(dHdl):
    with pytest.raises(ValueError, match="should be same as the length of series"):
        statistical_inefficiency(dHdl, series=dHdl[:10])


def test_series_none(dHdl):
    with pytest.warns(
        UserWarning,
        match="The series input is `None`, would not subsample according to statistical inefficiency.",
    ):
        statistical_inefficiency(dHdl, series=None)



"""Parsers for extracting alchemical data from `NAMD <http://www.ks.uiuc.edu/Research/namd/>`_ output files."""

from os.path import basename
from re import split

import numpy as np
import pandas as pd
from loguru import logger

from . import _init_attrs
from .util import anyopen
from ..postprocessors.units import R_kJmol, kJ2kcal

k_b = R_kJmol * kJ2kcal


def _filename_sort_key(s: str) -> list[int | str]:
    """Key for natural-sorting filenames, ignoring the path.

    This means that unlike with the standard Python sorted() function, "foo9" < "foo10".
    """

    return [int(t) if t.isdigit() else t.lower() for t in split(r"(\d+)", basename(s))]


def _get_lambdas(fep_files: str | list[str]) -> None | list[float]:
    """Retrieves all lambda values included in the FEP files provided.

    We have to do this in order to tolerate truncated and restarted fepout files.
    The IDWS lambda is not present at the termination of the window, presumably
    for backwards compatibility with ParseFEP and probably other things.

    For a given lambda1, there can be only one lambda2 and at most one lambda_idws.

    Parameters
    ----------
    fep_files: str or list of str
        Path(s) to fepout files to extract data from.

    Returns
    -------
    List of floats, or None if there is more than one lambda_idws for each lambda1.
    """

    lambda_fwd_map: dict[float, float] = {}
    lambda_bwd_map: dict[float, float] = {}
    is_ascending = set()
    endpoint_windows = []

    for fep_file in sorted(fep_files, key=_filename_sort_key):
        with anyopen(fep_file, "r") as f:  # type: ignore[arg-type]
            for line in f:
                line_split = line.strip().split()

                # We might not have a #NEW line so make the best guess
                if line_split[0] == "#NEW":
                    lambda1, lambda2 = float(line_split[6]), float(line_split[8])
                    lambda_idws = (
                        float(line_split[10]) if "LAMBDA_IDWS" in line_split else None
                    )
                elif line_split[0] == "#Free":
                    lambda1, lambda2, lambda_idws = (
                        float(line_split[7]),
                        float(line_split[8]),
                        None,
                    )
                else:
                    # We only care about lines with lambda values. No need to
                    # do all that other processing below for every line
                    continue  # pragma: no cover

                # Keep track of whether the lambda values are increasing or decreasing, so we can return
                # a sorted list of the lambdas in the correct order.
                # If it changes during parsing of this set of fepout files, then we know something is wrong

                # Keep track of endpoints separately since in IDWS runs there must be one of opposite direction
                if 0.0 in (lambda1, lambda2) or 1.0 in (lambda1, lambda2):
                    endpoint_windows.append((lambda1, lambda2))
                else:
                    # If the lambdas are equal then this doesn't represent an ascending window
                    if lambda2 != lambda1:
                        is_ascending.add(lambda2 > lambda1)
                    if lambda_idws is not None and lambda1 != lambda_idws:
                        is_ascending.add(lambda1 > lambda_idws)

                if len(is_ascending) > 1:
                    raise ValueError(
                        f"Lambda values change direction in {fep_file}, relative to the other files: {lambda1} -> {lambda2} (IDWS: {lambda_idws})"
                    )

                # Make sure the lambda2 values are consistent
                if lambda1 in lambda_fwd_map and lambda_fwd_map[lambda1] != lambda2:
                    logger.error(
                        f"fwd: lambda1 {lambda1} has lambda2 {lambda_fwd_map[lambda1]} in {fep_file} but it has already been {lambda2}"
                    )
                    raise ValueError(
                        "More than one lambda2 value for a particular lambda1"
                    )

                lambda_fwd_map[lambda1] = lambda2

                # Make sure the lambda_idws values are consistent
                if lambda_idws is not None:
                    if (
                        lambda1 in lambda_bwd_map
                        and lambda_bwd_map[lambda1] != lambda_idws
                    ):
                        logger.error(
                            f"bwd: lambda1 {lambda1} has lambda_idws {lambda_bwd_map[lambda1]} but it has already been {lambda_idws}"
                        )
                        raise ValueError(
                            "More than one lambda_idws value for a particular lambda1"
                        )
                    lambda_bwd_map[lambda1] = lambda_idws

    is_ascending = next(iter(is_ascending))  # type: ignore[arg-type]

    all_lambdas: set[float] = set()
    all_lambdas.update(lambda_fwd_map.keys())
    all_lambdas.update(lambda_fwd_map.values())
    all_lambdas.update(lambda_bwd_map.keys())
    all_lambdas.update(lambda_bwd_map.values())
    return list(sorted(all_lambdas, reverse=not is_ascending))


@_init_attrs
def extract_u_nk(fep_files: str | list[str], T: float) -> pd.DataFrame:
    """Return reduced potentials `u_nk` from NAMD fepout file(s).

    Parameters
    ----------
    fep_file : str or list of str
        Path to fepout file(s) to extract data from. These are sorted by filename,
        not including the path, prior to processing, using natural-sort. This way,
        filenames including numbers without leading zeros are handled intuitively.

        Windows may be split across files, or more than one window may be present
        in a given file. Windows without footer lines (which may be in a different
        file than the respective header lines) will raise an error. This means that
        while windows may have been interrupted and restarted, they must be
        complete. Lambda values are expected to increase or decrease monotonically,
        and match between header and footer of each window.

    T : float
        Temperature in Kelvin at which the simulation was sampled.

    Returns
    -------
    u_nk : DataFrame
        Potential energy for each alchemical state (k) for each frame (n).

    Note
    ----
    If the number of forward and backward samples in a given window are different,
    the extra sample(s) will be discarded. This is typically zero or one sample.


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.

    .. versionchanged:: 0.6.0
        Support for Interleaved Double-Wide Sampling files added, with various
        robustness checks.

        `fep_files` can now be a list of filenames.
    """
    beta = 1 / (k_b * T)

    # lists to get times and work values of each window
    win_ts: list[float] = []
    win_de: list[float] = []
    win_ts_back: list[float] = []
    win_de_back: list[float] = []

    # create dataframe for results
    u_nk = pd.DataFrame(columns=["time", "fep-lambda"])

    # boolean flag to parse data after equil time
    parsing = False

    if type(fep_files) is str:
        fep_files = [fep_files]

    # Extract the lambda values only from the fepouts
    all_lambdas = _get_lambdas(fep_files)
    # open and get data from fep file.
    # We sort the list of fep files in case some of them represent restarted windows.
    # The assumption is that they make sense in lexicographic order.
    # We keep track of which lambda window we're in, but since it can span multiple files,
    # only reset these variables here and after the end of each window
    lambda1_at_start, lambda2_at_start, lambda_idws_at_start = None, None, None

    for fep_file in sorted(fep_files, key=_filename_sort_key):
        # Note we have not set parsing=False because we could be continuing one window across
        # more than one fepout file
        with anyopen(fep_file, "r") as f:  # type: ignore[arg-type]
            has_idws = False
            for line in f:
                line_split = line.strip().split()
                # We don't know if IDWS was enabled just from the #Free line, and we might not have
                # a #NEW line in this file, so we have to check for the existence of FepE_back lines
                # We rely on short-circuit evaluation to avoid the string comparison most of the time
                if has_idws is False and line_split[0] == "FepE_back:":
                    has_idws = True

                # New window, get IDWS lambda if any
                # We keep track of lambdas from the #NEW line and if they disagree with the #Free line
                # within the same file, then complain. This can happen if truncated fepout files
                # are presented in the wrong order.
                if line_split[0] == "#NEW":
                    if parsing:
                        logger.error(
                            f"Window with lambda1: {lambda1_at_start} lambda2: {lambda2_at_start} lambda_idws: {lambda_idws_at_start} appears truncated"
                        )
                        logger.error(
                            f"because a new window was encountered in {fep_file} before the previous one finished."
                        )
                        raise ValueError("New window begun after truncated window")

                    lambda1_at_start, lambda2_at_start = (
                        float(line_split[6]),
                        float(line_split[8]),
                    )
                    lambda_idws_at_start = (
                        float(line_split[10]) if "LAMBDA_IDWS" in line_split else None
                    )
                    has_idws = True if lambda_idws_at_start is not None else False

                # this line marks end of window; dump data into dataframe
                if line_split[0] == "#Free":
                    # extract lambda values for finished window
                    # lambda1 = sampling lambda (row), lambda2 = comparison lambda (col)
                    lambda1 = float(line_split[7])
                    lambda2 = float(line_split[8])

                    # If the lambdas are not what we thought they would be, raise an exception to ensure the calculation
                    # fails. This can happen if fepouts where one window spans multiple fepouts are processed out of order
                    # NB: There is no way to tell if lambda_idws changed because it isn't in the '#Free' line that ends a window
                    if lambda1_at_start is not None and (lambda1, lambda2) != (
                        lambda1_at_start,
                        lambda2_at_start,
                    ):
                        logger.error(
                            f"Lambdas changed unexpectedly while processing {fep_file}"
                        )
                        logger.error(
                            f"l1, l2: {lambda1_at_start}, {lambda2_at_start} changed to {lambda1}, {lambda2}"
                        )
                        logger.error(line)
                        raise ValueError(
                            "Inconsistent lambda values within the same window"
                        )

                    # As we are at the end of a window, convert last window's work and times values to np arrays
                    # (with energy unit kT since they were kcal/mol in the fepouts)
                    win_de_arr = beta * np.asarray(win_de)  # dE values
                    win_ts_arr = np.asarray(win_ts)  # timesteps

                    # This handles the special case where there are IDWS energies but no lambda_idws value in the
                    # current .fepout file. This can happen when the NAMD firsttimestep is not 0, because NAMD only emits
                    # the '#NEW' line on timestep 0 for some reason. Perhaps the user ran minimize before dynamics,
                    # or this is a restarted run.
                    # We infer lambda_idws_at_start if it wasn't explictly included in this fepout.
                    # If lambdas are in ascending order, choose the one before it all_lambdas, and if descending, choose
                    # the one after. This happens "automatically" because the lambdas were returned already sorted
                    # in the correct direction by _get_lambdas().
                    # The "else" case is handled by the rest of this block, by default.
                    if has_idws and lambda_idws_at_start is None:
                        l1_idx = all_lambdas.index(lambda1)  # type: ignore[union-attr]
                        # Test for the highly pathological case where the first window is both incomplete and has IDWS
                        # data but no lambda_idws value.
                        if l1_idx == 0:
                            raise ValueError(
                                "IDWS data present in first window but lambda_idws not included; no way to infer the correct lambda_idws"
                            )
                        lambda_idws_at_start = all_lambdas[l1_idx - 1]  # type: ignore[index]
                        logger.warning(
                            f"Warning: {fep_file} has IDWS data but lambda_idws not included."
                        )
                        logger.warning(
                            f"         lambda1 = {lambda1}, lambda2 = {lambda2}; inferring lambda_idws to be {lambda_idws_at_start}"
                        )

                    if lambda_idws_at_start is not None:
                        # Mimic classic DWS data
                        # Arbitrarily match up fwd and bwd comparison energies on the same times
                        # truncate extra samples from whichever array is longer
                        win_de_back_arr = beta * np.asarray(win_de_back)
                        n = min(len(win_de_back_arr), len(win_de_arr))

                        tempDF = pd.DataFrame(
                            {
                                "time": win_ts_arr[:n],
                                "fep-lambda": np.full(n, lambda1),
                                lambda1: 0,
                                lambda2: win_de_arr[:n],
                                lambda_idws_at_start: win_de_back_arr[:n],
                            }
                        )
                        # print(f"{fep_file}: IDWS window {lambda1} {lambda2} {lambda_idws_at_start}")
                    else:
                        # print(f"{fep_file}: Forward-only window {lambda1} {lambda2}")
                        # create dataframe of times and work values
                        # this window's data goes in row LAMBDA1 and column LAMBDA2
                        tempDF = pd.DataFrame(
                            {
                                "time": win_ts_arr,
                                "fep-lambda": np.full(len(win_de_arr), lambda1),
                                lambda1: 0,
                                lambda2: win_de_arr,
                            }
                        )

                    # join the new window's df to existing df
                    u_nk = pd.concat([u_nk, tempDF], sort=False)

                    # reset values for next window of fepout file
                    win_de = []
                    win_ts = []
                    win_de_back = []
                    win_ts_back = []
                    parsing = False
                    has_idws = False
                    lambda1_at_start, lambda2_at_start, lambda_idws_at_start = (
                        None,
                        None,
                        None,
                    )

                # append work value from 'dE' column of fepout file
                if parsing:
                    if line_split[0] == "FepEnergy:":
                        win_de.append(float(line_split[6]))
                        win_ts.append(float(line_split[1]))
                    elif line_split[0] == "FepE_back:":
                        win_de_back.append(float(line_split[6]))
                        win_ts_back.append(float(line_split[1]))

                # Turn parsing on after line 'STARTING COLLECTION OF ENSEMBLE AVERAGE'
                if "#STARTING" in line_split:
                    parsing = True

    if len(win_de) != 0 or len(win_de_back) != 0:  # pragma: no cover
        logger.warning(
            'Trailing data without footer line ("#Free energy..."). Interrupted run?'
        )
        raise ValueError("Last window is truncated")

    if lambda2 in (0.0, 1.0):
        # this excludes the IDWS case where a dataframe already exists for both endpoints
        # create last dataframe for fep-lambda at last LAMBDA2
        tempDF = pd.DataFrame({"time": win_ts_arr, "fep-lambda": lambda2})
        u_nk = pd.concat([u_nk, tempDF], sort=True)

    u_nk.set_index(["time", "fep-lambda"], inplace=True)

    return u_nk


def extract(fep_files: str | list[str], T: float) -> dict[str, pd.DataFrame | None]:
    r"""Return reduced potentials `u_nk` and gradients `dH/dl`
    from NAMD fepout file(s).

    Parameters
    ----------
    fep_file : str or list of str
        Path to fepout file(s) to extract data from. These are sorted by filename,
        not including the path, prior to processing, using natural-sort. This way,
        filenames including numbers without leading zeros are handled intuitively.

        Windows may be split across files, or more than one window may be present
        in a given file. Windows without footer lines (which may be in a different
        file than the respective header lines) will raise an error. This means that
        while windows may have been interrupted and restarted, they must be
        complete. Lambda values are expected to increase or decrease monotonically,
        and match between header and footer of each window.

    T : float
        Temperature in Kelvin at which the simulation was sampled.

    Returns
    -------
    Dict
        A dictionary with keys of 'u_nk', which is a pandas DataFrame of
        potential energy for each alchemical state (k) for each frame (n).

    Note
    ----
    If the number of forward and backward samples in a given window are different,
    the extra sample(s) will be discarded. This is typically zero or one sample.


    .. versionadded:: 1.0.0
    """

    return {
        "u_nk": extract_u_nk(fep_files, T)
    }  # NOTE: maybe we should also have 'dHdl': None



"""Functions for Plotting the dF states.

To assess the quality of the free energy estimation, The dF between adjacent
lambda states can be plotted to assess the quality of the estimation.

The code for producing the dF states plot is modified based on
`Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.font_manager import FontProperties as FP
from matplotlib.figure import Figure
from typing import Any

from ..estimators import TI, BAR, MBAR
from ..postprocessors.units import get_unit_converter


def plot_dF_state(
    estimators: Any,
    labels: None | list[str] = None,
    colors: None | list[str] = None,
    units: None | str = None,
    orientation: str = "portrait",
    nb: int = 10,
) -> Figure:
    """Plot the dhdl of TI.

    Parameters
    ----------
    estimators : :class:`~alchemlyb.estimators` or list
        One or more :class:`~alchemlyb.estimators`, where the
        dhdl value will be taken from. For more than one estimators
        with more than one alchemical transformation, a list of list format
        is used.
    labels : List
        list of labels for labelling different estimators.
    colors : List
        list of colors for plotting different estimators.
    units : str
        The unit of the estimate. The default is `None`, which is to use the
        unit in the input. Setting this will change the output unit.
    orientation : string
        The orientation of the figure. Can be `portrait` or `landscape`
    nb : int
        Maximum number of dF states in one row in the `portrait` mode

    Returns
    -------
    matplotlib.figure.Figure
        An Figure with the dF states drawn.

    Note
    ----
    The code is taken and modified from
    `Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.


    .. versionchanged:: 1.0.0
        If no units is given, the `units` in the input will be used.

    .. versionchanged:: 0.5.0
        The `units` will be used to change the underlying data instead of only
        changing the figure legend.

    .. versionadded:: 0.4.0
    """
    try:
        len(estimators)
    except TypeError:
        estimators = [
            estimators,
        ]

    formatted_data = []
    for dhdl in estimators:
        try:
            len(dhdl)
            formatted_data.append(dhdl)
        except TypeError:
            formatted_data.append(
                [
                    dhdl,
                ]
            )

    if units is None:
        units = formatted_data[0][0].delta_f_.attrs["energy_unit"]

    estimators = formatted_data

    # Get the dF
    dF_list = []
    error_list = []
    max_length = 0
    convert = get_unit_converter(units)
    for dhdl_list in estimators:
        len_dF = sum([len(dhdl.delta_f_) - 1 for dhdl in dhdl_list])
        if len_dF > max_length:
            max_length = len_dF
        dF = []
        error = []
        for dhdl in dhdl_list:
            for i in range(len(dhdl.delta_f_) - 1):
                dF.append(convert(dhdl.delta_f_).iloc[i, i + 1])
                error.append(convert(dhdl.d_delta_f_).iloc[i, i + 1])

        dF_list.append(dF)
        error_list.append(error)

    # Get the determine orientation
    if orientation == "landscape":
        if max_length < 8:
            fig, ax = plt.subplots(figsize=(8, 6))
        else:
            fig, ax = plt.subplots(figsize=(max_length, 6))
        axs = [
            ax,
        ]
        xs = [
            np.arange(max_length),
        ]
    elif orientation == "portrait":
        if max_length < nb:
            xs = [
                np.arange(max_length),
            ]
            fig, ax = plt.subplots(figsize=(8, 6))
            axs = [
                ax,
            ]
        else:
            xs = np.array_split(np.arange(max_length), max_length / nb + 1)  # type: ignore[call-overload]
            fig, axs = plt.subplots(nrows=len(xs), figsize=(8, 6))
        mnb = max([len(i) for i in xs])
    else:
        raise ValueError(
            "Not recognising {}, only supports 'landscape' or 'portrait'.".format(
                orientation
            )
        )

    # Sort out the colors
    if colors is None:
        colors_dict = {
            "TI": "#C45AEC",
            "TI-CUBIC": "#33CC33",
            "DEXP": "#F87431",
            "IEXP": "#FF3030",
            "GINS": "#EAC117",
            "GDEL": "#347235",
            "BAR": "#6698FF",
            "UBAR": "#817339",
            "RBAR": "#C11B17",
            "MBAR": "#F9B7FF",
        }
        colors = []
        for dhdl in estimators:
            dhdl = dhdl[0]
            if isinstance(dhdl, TI):
                colors.append(colors_dict["TI"])
            elif isinstance(dhdl, BAR):
                colors.append(colors_dict["BAR"])
            elif isinstance(dhdl, MBAR):
                colors.append(colors_dict["MBAR"])
    else:
        if len(colors) >= len(estimators):
            pass
        else:
            raise ValueError(
                "Number of colors ({}) should be larger than the number of data ({})".format(
                    len(colors), len(estimators)
                )
            )

    # Sort out the labels
    if labels is None:
        labels = []
        for dhdl in estimators:
            dhdl = dhdl[0]
            if isinstance(dhdl, TI):
                labels.append("TI")
            elif isinstance(dhdl, BAR):
                labels.append("BAR")
            elif isinstance(dhdl, MBAR):
                labels.append("MBAR")
    else:
        if len(labels) == len(estimators):
            pass
        else:
            raise ValueError(
                "Length of labels ({}) should be the same as the number of data ({})".format(
                    len(labels), len(estimators)
                )
            )

    # Plot the figure
    width = 1.0 / (len(estimators) + 1)
    elw = 30 * width
    ndx = 1
    for x, ax in zip(xs, axs):
        lines: list[Any] = []
        for i, (dF, error) in enumerate(zip(dF_list, error_list)):
            y = [dF[j] for j in x]
            ye = [error[j] for j in x]
            if orientation == "landscape":
                lw = 0.1 * elw
            elif orientation == "portrait":
                lw = 0.05 * elw
            line = ax.bar(
                x + len(lines) * width,
                y,
                width,
                color=colors[i],
                yerr=ye,
                lw=lw,
                error_kw=dict(elinewidth=elw, ecolor="black", capsize=0.5 * elw),
            )
            lines += (line[0],)
        for dir in ["left", "right", "top", "bottom"]:
            if dir == "left":
                ax.yaxis.set_ticks_position(dir)  # type: ignore[arg-type]
            else:
                ax.spines[dir].set_color("none")

        if orientation == "landscape":
            plt.yticks(fontsize=8)
            ax.set_xlim(x[0] - width, x[-1] + len(lines) * width)
            plt.xticks(
                x + 0.5 * width * len(estimators),
                tuple([f"{i}--{i + 1}" for i in x]),
                fontsize=8,
            )
        elif orientation == "portrait":
            plt.yticks(fontsize=10)
            ax.xaxis.set_ticks([])
            for i in x + 0.5 * width * len(estimators):
                ax.annotate(
                    r"$\mathrm{%d-%d}$" % (i, i + 1),
                    xy=(i, 0),
                    xycoords=("data", "axes fraction"),
                    xytext=(0, -2),
                    size=10,
                    textcoords="offset points",
                    va="top",
                    ha="center",
                )
            ax.set_xlim(x[0] - width, x[-1] + len(lines) * width + (mnb - len(x)))
        ndx += 1
    x = np.arange(max_length)

    ax = plt.gca()

    for tick in ax.get_xticklines():
        tick.set_visible(False)
    if orientation == "landscape":
        leg = plt.legend(lines, labels, loc=3, ncol=2, prop=FP(size=10), fancybox=True)
        plt.title("The free energy change breakdown", fontsize=12)
        plt.xlabel("States", fontsize=12, color="#151B54")
        plt.ylabel(r"$\Delta G$ ({})".format(units), fontsize=12, color="#151B54")
    elif orientation == "portrait":
        leg = ax.legend(
            lines,
            labels,
            loc=0,
            ncol=2,
            prop=FP(size=8),
            title=r"$\Delta G$ ({})".format(units) + r"$\mathit{vs.}$ lambda pair",
            fancybox=True,
        )

    leg.get_frame().set_alpha(0.5)
    return fig



import numpy as np
import pandas as pd
import pytest

from alchemlyb import concat
from alchemlyb.convergence import (
    forward_backward_convergence,
    fwdrev_cumavg_Rc,
    A_c,
    block_average,
)
from alchemlyb.convergence.convergence import _cummean


def test_convergence_ti(gmx_benzene_Coulomb_dHdl):
    convergence = forward_backward_convergence(gmx_benzene_Coulomb_dHdl, "TI")
    assert convergence.shape == (10, 5)

    assert convergence.loc[0, "Forward"] == pytest.approx(3.07, 0.01)
    assert convergence.loc[0, "Backward"] == pytest.approx(3.11, 0.01)
    assert convergence.loc[9, "Forward"] == pytest.approx(3.09, 0.01)
    assert convergence.loc[9, "Backward"] == pytest.approx(3.09, 0.01)


@pytest.mark.parametrize("estimator", ["MBAR", "BAR"])
def test_convergence_fep(gmx_benzene_Coulomb_u_nk, estimator):
    convergence = forward_backward_convergence(gmx_benzene_Coulomb_u_nk, estimator)
    assert convergence.shape == (10, 5)
    assert convergence.loc[0, "Forward"] == pytest.approx(3.02, 0.01)
    assert convergence.loc[0, "Backward"] == pytest.approx(3.06, 0.01)
    assert convergence.loc[9, "Forward"] == pytest.approx(3.05, 0.01)
    assert convergence.loc[9, "Backward"] == pytest.approx(3.04, 0.01)


def test_block_average_ti(gmx_benzene_Coulomb_dHdl):
    df_avg = block_average(gmx_benzene_Coulomb_dHdl, "TI")
    assert df_avg.shape == (9, 2)
    assert df_avg.loc[1, "FE"] == pytest.approx(3.18, 0.01)
    assert df_avg.loc[1, "FE_Error"] == pytest.approx(0.07, 0.1)
    assert df_avg.loc[8, "FE"] == pytest.approx(3.15, 0.01)
    assert df_avg.loc[8, "FE_Error"] == pytest.approx(0.07, 0.1)


@pytest.mark.parametrize("estimator", ["DUMMY"])
def test_block_average_error_1(gmx_ABFE_complex_u_nk, estimator):
    with pytest.raises(ValueError, match=r"Estimator DUMMY is not available .*"):
        _ = block_average(gmx_ABFE_complex_u_nk, estimator)


@pytest.mark.parametrize("estimator", ["MBAR"])
def test_block_average_error_2_mbar(gmx_ABFE_complex_u_nk, estimator):
    df_list = gmx_ABFE_complex_u_nk[10:15]
    with pytest.raises(
        ValueError,
        match=r"Provided DataFrame, df_list\[0\] has more than one lambda value in df.index\[0\]",
    ):
        _ = block_average([concat(df_list)], estimator)

    df_list = gmx_ABFE_complex_u_nk[14:17]
    with pytest.raises(
        ValueError,
        match=r"Provided DataFrame, df_list\[0\] has more than one lambda value in df.index\[1\]",
    ):
        _ = block_average([concat(df_list)], estimator)


@pytest.mark.parametrize("estimator", ["BAR"])
def test_block_average_error_2_bar(gmx_ABFE_complex_u_nk, estimator):
    df_list = gmx_ABFE_complex_u_nk[10:13]
    with pytest.raises(
        ValueError,
        match=r"Restrict to two DataFrames, one with a fep-lambda value .*",
    ):
        _ = block_average(df_list, estimator)

    df_list = gmx_ABFE_complex_u_nk[14:17]
    with pytest.raises(
        ValueError,
        match=r"Restrict to two DataFrames, one with a fep-lambda value .*",
    ):
        _ = block_average(df_list, estimator)


@pytest.mark.parametrize("estimator", ["BAR"])
def test_block_average_error_3_bar(gmx_ABFE_complex_u_nk, estimator):
    # Test if lambda state column representing one of the two lambda
    # states in the df indices is missing from *both* dfs.
    df_list = gmx_ABFE_complex_u_nk[10:12]
    state1 = list(set(x[1:] for x in df_list[0].index))[0]
    df_list[0] = df_list[0].drop(state1, axis=1)
    df_list[1] = df_list[1].drop(state1, axis=1)
    with pytest.raises(
        ValueError,
        match=r"Indexed lambda state, .*",
    ):
        _ = block_average(df_list, estimator)


@pytest.mark.parametrize("estimator", ["BAR"])
def test_block_average_error_4_bar(gmx_ABFE_complex_u_nk, estimator):
    # Test if lambda state column representing one of the two lambda
    # states in the df indices is missing from *one* dfs.
    df_list = gmx_ABFE_complex_u_nk[10:12]
    state1 = list(set(x[1:] for x in df_list[0].index))[0]
    df_list[0] = df_list[0].drop(state1, axis=1)
    with pytest.raises(
        ValueError,
        match=r"u_nk does not contain energies computed between any adjacent .*",
    ):
        _ = block_average(df_list, estimator)


@pytest.mark.parametrize("estimator", ["BAR"])
def test_block_average_bar(gmx_ABFE_complex_u_nk, estimator):
    df_avg = block_average(gmx_ABFE_complex_u_nk[10:12], estimator)
    assert df_avg.shape == (9, 2)
    assert df_avg.loc[0, "FE"] == pytest.approx(3.701, 0.01)
    assert np.isnan(df_avg.loc[0, "FE_Error"])
    assert df_avg.loc[8, "FE"] == pytest.approx(3.603, 0.01)
    assert np.isnan(df_avg.loc[8, "FE_Error"])

    df_list = gmx_ABFE_complex_u_nk[14:16]
    df_list[-1] = df_list[-1].iloc[:-2]
    df_avg = block_average(df_list, estimator)
    assert df_avg.shape == (9, 2)
    assert df_avg.loc[0, "FE"] == pytest.approx(0.651, 0.01)
    assert np.isnan(df_avg.loc[0, "FE_Error"])
    assert df_avg.loc[8, "FE"] == pytest.approx(0.901, 0.01)
    assert np.isnan(df_avg.loc[8, "FE_Error"])


@pytest.mark.parametrize("estimator", ["MBAR"])
def test_block_average_mbar(gmx_benzene_Coulomb_u_nk, estimator):
    df_avg = block_average([gmx_benzene_Coulomb_u_nk[0]], estimator)
    assert df_avg.shape == (9, 2)
    assert df_avg.loc[0, "FE"] == pytest.approx(3.41, 0.01)
    assert df_avg.loc[0, "FE_Error"] == pytest.approx(0.22, 0.01)
    assert df_avg.loc[8, "FE"] == pytest.approx(2.83, 0.01)
    assert df_avg.loc[8, "FE_Error"] == pytest.approx(0.33, 0.01)


def test_convergence_wrong_estimator(gmx_benzene_Coulomb_dHdl):
    with pytest.raises(ValueError, match="is not available in"):
        forward_backward_convergence(gmx_benzene_Coulomb_dHdl, "WWW")


def test_convergence_wrong_cases(gmx_benzene_Coulomb_u_nk):
    with pytest.raises(ValueError, match="is not available in"):
        forward_backward_convergence(gmx_benzene_Coulomb_u_nk, "mbar")


def test_convergence_bootstrap(gmx_benzene_Coulomb_u_nk, caplog):
    normal_c = forward_backward_convergence(gmx_benzene_Coulomb_u_nk, "MBAR", num=2)
    bootstrap_c = forward_backward_convergence(
        gmx_benzene_Coulomb_u_nk, "MBAR", error_tol=0.01, num=2
    )
    assert "use bootstrap error instead." in caplog.text
    assert (bootstrap_c["Forward_Error"] != normal_c["Forward_Error"]).all()


def test_convergence_method(gmx_benzene_Coulomb_u_nk):
    convergence = forward_backward_convergence(
        gmx_benzene_Coulomb_u_nk, "MBAR", num=2, method="adaptive"
    )
    assert len(convergence) == 2


@pytest.mark.parametrize("estimator", ["MBAR"])
def test_forward_backward_convergence_mbar(gmx_ABFE_complex_u_nk, estimator):
    df_list = gmx_ABFE_complex_u_nk[10:15]
    with pytest.raises(
        ValueError,
        match=r"Provided DataFrame, df_list\[0\] has more than one lambda value in df.index\[0\]",
    ):
        _ = forward_backward_convergence([concat(df_list)], estimator)

    df_list = gmx_ABFE_complex_u_nk[14:17]
    with pytest.raises(
        ValueError,
        match=r"Provided DataFrame, df_list\[0\] has more than one lambda value in df.index\[1\]",
    ):
        _ = forward_backward_convergence([concat(df_list)], estimator)


def test_cummean_short():
    """Test the case where the input is shorter than the expected output"""
    value = _cummean(np.empty(10), 100)
    assert len(value) == 10


def test_cummean_long():
    """Test the case where the input is longer than the expected output"""
    value = _cummean(np.empty(20), 10)
    assert len(value) == 10


def test_cummean_long_none_integter():
    """Test the case where the input is not a integer multiple of the expected output"""
    value = _cummean(np.empty(25), 10)
    assert len(value) == 10


def test_R_c_converged():
    data = pd.Series(data=[0] * 100)
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value, running_average = fwdrev_cumavg_Rc(data)
    np.testing.assert_allclose(value, 0.0)


def test_R_c_notconverged():
    data = pd.Series(data=range(21))
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value, running_average = fwdrev_cumavg_Rc(data, tol=0.1, precision=0.05)
    np.testing.assert_allclose(value, 1.0)


def test_R_c_real():
    data = pd.Series(data=np.hstack((range(10), [4.5] * 10)))
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value, running_average = fwdrev_cumavg_Rc(data, tol=2.0)
    np.testing.assert_allclose(value, 0.35)


def test_A_c_real():
    data = pd.Series(data=np.hstack((range(10), [4.5] * 10)))
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value = A_c([data] * 2, tol=2.0)
    np.testing.assert_allclose(value, 0.65)
